#整个程序运行前将会先调用这个函数，可以不写
def init():
	from tuack import base
	#base.conf里面是预先读取好的conf.yaml数据，并对其进行了一些预处理
	#注意这段代码会【覆盖】data字段
	base.conf['data'] = []
	#这里是某种生成data参数的代码
	for i in range(1, 20):
		base.conf['data'].append({
			'cases' : [f"{i}-{j + 1}" for j in range(round(i ** 1.1))],
			'args' : {'n' : 20, 'm' : i}
		})
	#修改完conf以后需要保存
	base.save_json(base.conf)

from tuack import dmk
#如果使用random库，dmk会自动保存第一次运行的种子信息，之后运行会生成相同的数据
#如果需要重新随机，删除random_state即可
from random import randint

#生成数据的函数，参数分别是输出文件对象、参数获取器、上下文信息
def gen(fin, args, cont):
	# cont中有一些上下文信息
	dmk.log.debug(f"当前数据类型：{cont['folder']}")
	n = args['n']	#自动查找参数，优先查找测试点中的，找不到找全局，再找不到则是None
	m = args['m']
	tp = (cont['case'] if type(cont['case']) == int else int(cont['case'][-1])) & 1
	#往fin中输出即可
	fin.write(f"{n} {m}\n")
	for i in range(m - tp):
		fin.write(f"{randint(1, n)} {randint(1, n)}\n")
	if tp:
		fin.write('1 1\n')



# -*- coding: utf-8 -*-

from __future__ import print_function
import os, uuid
import re
import sys
import json
import datetime
import shutil
import subprocess
import time
import signal
import zipfile
from multiprocessing import Process, Queue
from functools import wraps
from threading import Timer
import platform
from . import base
from .base import log, pjoin
import requests
import traceback
from requests_toolbelt.multipart.encoder import MultipartEncoder, MultipartEncoderMonitor

def lemon(conf = None):
	py_version = base.python_version[0]
	if py_version == 2:
		base.check_install('pyside')
		from PySide import QtCore
	if not conf:
		if base.conf.folder == 'problem':
			raise Exception('Can\'t dump a single problem to lemon, try to dump a day or a contest.')
		base.remkdir('lemon')
		if base.conf.folder == 'day':
			lemon(base.conf)
		else:
			for day in base.days():
				os.makedirs(base.pjoin('lemon', day.route))
				lemon(day)
		return
	log.info(u'导出lemon工程：%s' % conf.route)
	os.makedirs(base.pjoin('lemon', conf.route, 'data'))
	os.makedirs(base.pjoin('lemon', conf.route, 'source'))
	if py_version == 2:
		jzp_magic = 0x20111127
		zlib_magic = 0x185E
		import zlib
		obuff = QtCore.QByteArray()
		ost = QtCore.QDataStream(obuff, QtCore.QIODevice.WriteOnly)
		ost.writeQString(conf['name'])
		probs = list(conf.probs())
		ost.writeInt32(len(probs))
		for prob in probs:
			log.info(u'导出lemon题目：%s' % prob.route)
			ost.writeQString(base.tr(prob['title']))
			ost.writeQString(prob['name'])
			ost.writeQString(prob['name'] + '.in')
			ost.writeQString(prob['name'] + '.out')
			ost.writeBool(False)
			ost.writeBool(False)
			ost.writeInt32(1 if prob['type'] == 'output' else 0)
			ost.writeInt32(1)		# Judge Type TODO: What if there is spj (code = 4)
			ost.writeQString('--ignore-space-change --text --brief')
			ost.writeInt32(3)		# real precision (float number error bound)
			ost.writeQString('')	# spj route TODO: What if there is spj
			ost.writeInt32(len([i for i in prob.get('compile', {}) if i in {'cpp', 'c', 'pas'}]))
			for key, val in prob.get('compile', {}).items():
				try:
					ost.writeQString({
						'cpp' : 'g++',
						'c' : 'gcc',
						'pas' : 'fpc'
					}[key])
					ost.writeQString(val)
				except:
					pass
			ost.writeQString('out')
			if prob.packed:
				ost.writeInt32(len(prob['data']))
				for datum in prob['data']:
					ost.writeInt32(datum['score'])
					ost.writeInt32(datum.get('time limit', prob.get('time limit', 0)) * 1000)
					ost.writeInt32(base.Memory(datum.get('memory limit', prob.get('memory limit', '0 B'))).MB)
					tc = datum['cases']
					ost.writeInt32(len(tc))
					for c in tc:
						ost.writeQString(base.pjoin(prob['name'], str(c) + '.in'))
					ost.writeInt32(len(tc))
					for c in tc:
						ost.writeQString(base.pjoin(prob['name'], str(c) + '.ans'))
			else:
				score = (100. / len(prob.test_cases) if len(prob.test_cases) > 0 else 0.)
				if 100 % len(prob.test_cases) != 0:
					log.warning('测试点个数不是100的约数，总分将不是整百。')
				ost.writeInt32(len(prob.test_cases))
				for c in prob.test_cases:
					ost.writeInt32(int(score + .5))
					ost.writeInt32(int(prob['time limit'] * 1000 + .5))
					ost.writeInt32(int(prob.memory_limit().MB + .5))
					ost.writeInt32(1)
					ost.writeQString(base.pjoin(prob['name'], str(c) + '.in'))
					ost.writeInt32(1)
					ost.writeQString(base.pjoin(prob['name'], str(c) + '.ans'))
			
			sp = list(prob.route.split('/'))
			target = ['lemon'] + sp[:-1] + ['data', sp[-1]]
			shutil.copytree(base.pjoin(prob.path, 'data'), base.pjoin(*target))

		ss = lambda s : str(s) if py_version == 2 else s
		compressed = QtCore.QByteArray(zlib.compress(ss(obuff)))
		obuff = QtCore.QByteArray()
		ost = QtCore.QDataStream(obuff, QtCore.QIODevice.WriteOnly)
		ost.writeUInt32(zlib_magic)
		ost.writeRawData(ss(compressed))
		file_ = QtCore.QFile(base.pjoin('lemon', conf.route, conf['name'] + '.cdf'))
		file_.open(QtCore.QIODevice.WriteOnly)
		ofs = QtCore.QDataStream(file_)
		ofs.writeUInt32(jzp_magic)
		ofs.writeUInt16(QtCore.qChecksum(ss(obuff), len(obuff)))
		ofs.writeUInt32(len(obuff))
		ofs.writeRawData(ss(obuff))
		file_.close()
	else:
		probs = list(conf.probs())
		tasks = []
		for prob in probs:
			log.info(u'导出lemon题目：%s' % prob.route)
			if prob.packed:
				cases = [
					{
						"fullScore": int(datum['score'] + .5),
						"timeLimit": int(datum.get('time limit', prob.get('time limit', 0)) * 1000 + .5),
						"memoryLimit": int(base.Memory(datum.get('memory limit', prob.get('memory limit', '0 B'))).MB + .5),
						"inputFiles": [
							base.pjoin(prob['name'], str(c) + '.in') \
							for c in datum['cases']
						],
						"outputFiles": [
							base.pjoin(prob['name'], str(c) + '.ans') \
							for c in datum['cases']
						]
					} for datum in prob['data']
				]
			else:
				score = (100. / len(prob.test_cases) if len(prob.test_cases) > 0 else 0.)
				if 100 % len(prob.test_cases) != 0:
					log.warning('测试点个数不是100的约数，总分将不是整百。')
				cases = [
					{
						"fullScore": int(score + .5),
						"timeLimit": int(prob['time limit'] * 1000 + .5),
						"memoryLimit": int(prob.memory_limit().MB + .5),
						"inputFiles": [
							base.pjoin(prob['name'], str(c) + '.in')
						],
						"outputFiles": [
							base.pjoin(prob['name'], str(c) + '.ans')
						]
					} for c in prob.test_cases
				]
			task = {
				'answerFileExtension' : 'out',
				"comparisonMode": 1,
				"diffArguments": "--ignore-space-change --text --brief",
				"inputFileName": prob['name'] + '.in',
				"outputFileName": prob['name'] + '.out',
				"problemTitle": base.tr(prob['title']),
				"realPrecision": 3,		#float number error bound
				"sourceFileName": prob['name'],
				"specialJudge": "",
				"standardInputCheck": False,
				"standardOutputCheck": False,
				"subFolderCheck": False,
				"taskType": 1 if prob['type'] == 'output' else 0,
				"compilerConfiguration": {
					{
						'cpp' : 'g++',
						'c' : 'gcc',
						'pas' : 'fpc',
						'py' : 'python',
						'java' : 'javac'
					}[i] : 'default' for i in prob.get('compile', {}) if i in {'cpp', 'c', 'pas', 'py', 'java'}
				},
				"testCases" : cases
			}
			sp = list(prob.route.split('/'))
			target = ['lemon'] + sp[:-1] + ['data', sp[-1]]
			shutil.copytree(base.pjoin(prob.path, 'data'), base.pjoin(*target))
			if os.path.isdir(base.pjoin(prob.path, 'data', 'chk')) and os.path.isfile(base.pjoin(prob.path, 'data', 'chk', 'chk.cpp')):
				log.info('发现chk，尝试编译。')
				os.system("g++ %s -o %s -O2 -std=c++17" % (
					base.pjoin(prob.path, 'data', 'chk', 'chk.cpp'),
					base.pjoin(*(target + ['chk.exe']))
				))
				task['comparisonMode'] = 4
				task['specialJudge'] = base.pjoin(prob['name'], 'chk.exe')
			tasks.append(task)
		open(base.pjoin('lemon', conf.route, conf['name'] + '.cdf'), 'wb').write(json.dumps({
			"contestTitle" : conf['name'],
			"contestants" : [],
			"tasks" : tasks
		}).encode('utf-8'))
	
	base.run_r(base.unix2dos, base.pjoin('lemon', conf.route, 'data'))
	
	if base.do_zip:
		import zipfile
		with zipfile.ZipFile(base.pjoin('lemon', conf.route) + '.zip', 'w') as z:
			base.run_r(lambda path : z.write(path), base.pjoin('lemon', conf.route))
	log.warning(u'目前lemon的编译选项是写在注册表里的，暂时没有实现该功能，请手工配置。')

def arbiter_main(conf = None,daynum = 0):
	def arbiter_info(info, filename):
		with open(filename,'wb') as ofile:
			for key, val in info.items():
				ofile.write(('%s%s\n' % (key, val)).encode('utf-8'))
	if not conf:
		log.info('makedirs')
		if not os.path.exists('arbiter'):
			base.remkdir('arbiter')
		base.remkdir(pjoin('arbiter', 'main'))
		os.makedirs(base.pjoin('arbiter', 'main','data'))
		os.makedirs(base.pjoin('arbiter', 'main','final'))
		os.makedirs(base.pjoin('arbiter', 'main','players'))
		os.makedirs(base.pjoin('arbiter', 'main','result'))
		os.makedirs(base.pjoin('arbiter', 'main','filter'))
		os.makedirs(base.pjoin('arbiter', 'main','tmp'))
		if base.conf.folder == 'problem':
			raise Exception('Can\'t dump a single problem to arbiter, try to dump a day or a contest.')
		if base.conf.folder == 'day':
			arbiter(base.days())
		else:
			for idx, day in enumerate(base.days(), start = 1):
				os.makedirs(base.pjoin('arbiter', 'main','players','day%d' % idx))
				os.makedirs(base.pjoin('arbiter', 'main','result','day%d' % idx))
				arbiter_main(day,idx)
			log.info('dos2unix')
			base.run_r(base.dos2unix, base.pjoin('arbiter', 'main', 'data'))
			shutil.copytree(base.pjoin('arbiter', 'main','data'),base.pjoin('arbiter', 'main','evaldata'))	#这里也不能直接copy，见下面的处理方式
			cfg = {}
			cfg['NAME='] = base.conf['name']
			cfg['DAYNUM='] = idx
			cfg['ENV='] = 'env.info'
			cfg['PLAYER='] = 'player.info'
			cfg['TEAM='] = 'team.info'
			cfg['MISC='] = 'misc.info'
			arbiter_info(cfg,base.pjoin('arbiter', 'main','setup.cfg'))
			team = {}
			arbiter_info(team,base.pjoin('arbiter', 'main','team.info'))
			'''arbiter_info(userlist,base.pjoin('arbiter','player.info'))'''
		return
	dayinfo = {}
	dayinfo['NAME='] = u'第'+str(daynum)+u'场'+u'--机试'
	dayinfo['PLAYERDIR='] = ''
	dayinfo['CASEDIR='] = ''
	dayinfo['BASESCORE='] = 0
	dayinfo['TASKNUM='] = len(conf['subdir'])
	arbiter_info(dayinfo,base.pjoin('arbiter', 'main', 'day%d.info' % daynum))
	for probnum, prob in enumerate(conf.sub, start = 1):
		log.info(prob['name'])
		probinfo = {}
		probinfo['TITLE='] = ''
		probinfo['NAME='] = prob['name']
		probinfo['RUN='] = ''
		probinfo['INFILESUFFIX='] = 'in'
		probinfo['ANSFILESUFFIX='] = 'ans'
		probinfo['PLUG='] = prob['name']+'_e'
		if prob['type'] == 'program':
			probinfo['TYPE='] = 'SOURCE'
		else:
			log.warning(u'暂时只支持非交互式程序题。')
		probinfo['LIMIT='] = prob['time limit']
		probinfo['MEMLIMITS='] = int(prob.memory_limit().MB)
		probinfo['SAMPLES='] = len(prob.test_cases)
		score_per_case = 100 // len(prob.test_cases)
		if not prob.packed and score_per_case * len(prob.test_cases) != 100:
			log.warning(u'测试点数量不是100的约数，分数无法均分为整数。')
		probinfo['CCL=c@gcc'] = ' -o %o %i ' + prob['compile'].get('c', '')
		probinfo['CCL=cpp@g++'] = ' -o %o %i ' + prob['compile'].get('cpp', '')
		probinfo['CCL=pas@fpc'] = ' %i ' + prob['compile'].get('pas', '')
		idx = 0
		for datum in prob.data:
			if prob.packed and len(datum['cases']) > 1:
				log.warning(u'Arbiter不支持打包评测，将把该包得分均分给包中各测试点，无法整除时总分将不正确。')
			for case in datum['cases']:
			#for idx, case in enumerate(prob.test_cases, start = 1):
				idx += 1
				'''print('copyfile %s'%base.pjoin(prob.path,'data',case+'.in'))'''
				shutil.copy(
					base.pjoin(prob.path,'data',str(case)+'.in'),
					base.pjoin('arbiter', 'main','data',prob['name']+str(idx)+'.in')
				)
				'''print('copyfile %s'%base.pjoin(prob.path,'data',case+'.ans'))'''
				shutil.copy(
					base.pjoin(prob.path,'data',str(case)+'.ans'),
					base.pjoin('arbiter', 'main','data',prob['name']+str(idx)+'.ans')
				)
				if prob.packed:
					probinfo['MARK='+str(idx)+'@'] = str(int(datum['score'] / len(datum['cases'])))
				else:
					probinfo['MARK='+str(idx)+'@'] = str(int(score_per_case))
		'''for idx, userdir in enumerate(prob['users'],start = 1):
			for idx2, code in enumerate(prob['users'][userdir],start = 1):
				dirname = prob['name']+'-'+str(idx)+str(idx2)
				codename = userdir + code
				tmplist = prob['users'][userdir][code].split('/')
				codedir = base.pjoin(prob.path,*tmplist)
				os.makedirs(base.pjoin('arbiter','players',conf['name'],dirname,prob['name']))
				shutil.copy(codedir,base.pjoin('arbiter','players',conf['name'],dirname,prob['name']))
				userlist[dirname + '@'] =  codename'''
		
		if os.path.isdir(base.pjoin(prob.path, 'data', 'chk')) and os.path.isfile(base.pjoin(prob.path, 'data', 'chk', 'chk.cpp')):
			log.info('发现chk，尝试编译。')
			os.system("g++ %s -o %s -O2 -std=c++17" % (
				base.pjoin(prob.path, 'data', 'chk', 'chk.cpp'),
				base.pjoin('arbiter', 'main','filter',prob['name']+'_e')
			))
		else:
			shutil.copy(base.pjoin(base.path,'sample','arbiter_e.sample'),base.pjoin('arbiter', 'main','filter',prob['name']+'_e'))
		arbiter_info(probinfo,base.pjoin('arbiter', 'main','task'+str(daynum)+'_'+str(probnum)+'.info'))

def arbiter_down(conf = None):
	if not conf:
		conf = base.conf
		if not os.path.exists('arbiter'):
			base.remkdir('arbiter')
		base.remkdir(pjoin('arbiter', 'down'))
	if conf.folder == 'problem':
		raise Exception('Can\'t dump a single problem to arbiter, try to dump a day or a contest.')
	if conf.folder == 'contest':
		for idx, day in enumerate(base.days(), start = 1):
			os.makedirs(base.pjoin(pjoin('arbiter', 'down'), day['name']))
			arbiter_down(day)
		return
	for prob in conf.probs():
		log.info(prob.route)
		os.makedirs(base.pjoin(pjoin('arbiter', 'down'), prob.route))
		vis = set()
		for idx, case in enumerate(prob.sample_cases, start = 1):
			for suf in ('.in', '.ans'):
				s_name = base.pjoin(prob.path, 'down', str(case) + suf)
				t_name = base.pjoin('arbiter', 'down', prob.route, prob['name'] + str(idx) + suf)
				base.shutil_copy(s_name, t_name)
				vis.add(str(case) + suf)
		for fname in os.listdir(base.pjoin(prob.path, 'down')):
			if fname not in vis:
				log.info(f"找到了不在down列表中的文件{fname}")
				base.shutil_copy(
					base.pjoin(prob.path, 'down', fname),
					base.pjoin('arbiter', 'down', prob.route, fname)
				)
	log.info('dos2unix')
	base.run_r(base.dos2unix, base.pjoin(pjoin('arbiter', 'down', conf.route)))

def arbiter():
	base.remkdir('arbiter')
	arbiter_main()
	arbiter_down()

def tsinsen_oj():
	import random
	if type(base.conf) != base.Problem:
		log.error(u'只能转换一道题目，请到相应题目目录下运行')
		return
	hash_ch = list(map(
		chr,
		[ord('a') + i for i in range(26)] + \
		[ord('A') + i for i in range(26)] + \
		[ord('0') + i for i in range(10)]
	))
	hash = lambda l = 10 : ''.join([hash_ch[random.randint(0, len(hash_ch) - 1)] for i in range(l)])
	if not os.path.exists('tsinsen-oj'):
		os.makedirs('tsinsen-oj')
	for day in base.days():
		p = base.pjoin('tsinsen-oj', day.route)
		if not os.path.exists(p):
			os.makedirs(p)
	for prob in base.probs():
		p = base.pjoin('tsinsen-oj', prob.route)
		if not os.path.exists(p):
			os.makedirs(p)
		if os.path.exists(base.pjoin(prob.path, 'down')):
			with zipfile.ZipFile(base.pjoin('tsinsen-oj', prob.route, 'down.zip'), 'w') as z:
				base.run_r(lambda path : z.write(path), base.pjoin(prob.path, 'down'))
			prob.tsinsen_down = hash(8)
		else:
			prob.tsinsen_down = None
		files = {}
		path = base.pjoin(prob.path, 'resources')
		if os.path.exists(path):
			base.run_r(lambda p : files.__setitem__(p[len(path) + 1:], hash(8)), path)
		prob.tsinsen_files = files
	if base.do_render:
		from . import ren
		tmp = base.start_file
		base.start_file = False
		ren.Html('tsinsen-oj').run()
		base.start_file = tmp
	else:
		log.warning(u'如果你使用了文件，不重新渲染题面会导致tsinsen的文件失效。')
	
	read_file = lambda path : open(path, 'rb').read().decode('utf-8')
	
	Title = lambda : prob.tr('title')
	CheckPoint = lambda : prob['key words'] if 'key words' in prob else ''
	TestMethod = lambda : 'DEFAULT'
	InputFileName = lambda : ''
	OutputFileName = lambda : ''
	TimeLimit = lambda : '%.1fs' % prob['time limit']
	MemoryLimit = lambda : '%.1fMB' % prob.memory_limit().MB
	
	def Checkers():
		path = base.pjoin('data', 'chk', 'chk.cpp')
		if os.path.exists(path):
			return open(path, 'rb').read().decode('utf-8')
		else:
			return '\n'
	def Description():
		path = pjoin('statements', 'tsinsen-oj', prob.route if prob.route != '' else prob['name']) + '.html'
		if not os.path.exists(path):
			log.warning(u'找不到题面，你可能需要自己手工添加题面。')
			return ''
		header = u'<p>下载目录：<img src="/RequireFile.do?fid=%s" alt="另存为图片下载" />（另存为图片下载）</p>\n' % prob.tsinsen_down
		return header + read_file(path)
	def Solution():
		for user, codes in prob['users'].items():
			for name, path in codes.items():
				if prob.expect(user, name, 100):
					return read_file(path['path'])
		for user, codes in prob['users'].items():
			for name, path in codes.items():
				return read_file(path)
		return ''
	def add_shell(name, func):
		h = hash()
		ret = '%s=\n' % name
		ret += '======================%s\n' % h
		res = func()
		if not res.endswith('\n'):
			res += '\n'
		ret += res
		ret += '======================%s\n' % h
		return ret.encode('utf-8')
	def to_base64(path):
		import base64
		s = base64.b64encode(open(path, 'rb').read()).decode('utf-8')
		return '\n'.join(s[pos:pos+76] for pos in range(0, len(s), 76))
	tokens = [
		'Title', 'CheckPoint', 'Checkers', 'TestMethod', 'Description',
		'InputFileName', 'OutputFileName', 'Solution', 'TimeLimit', 'MemoryLimit'
	]
	for prob in base.probs():
		result_file = base.pjoin('tsinsen-oj', prob.route) + '.txt'
		with open(result_file, 'wb') as f:
			for token in tokens:
				f.write(add_shell(token, eval(token)))
			if prob.packed:
				log.warning(u'清橙OJ不支持打包评测和指定测试点分值，直接将所有测试点视为相同。')
			for datum in prob.test_cases:
				f.write(add_shell('InData', lambda : read_file(base.pjoin('data', datum + '.in'))))
				f.write(add_shell('OutData', lambda : read_file(base.pjoin('data', datum + '.ans'))))
			if prob.tsinsen_down:
				f.write(add_shell('FileName', lambda : 'down.zip'))
				f.write(add_shell('File(%s)' % prob.tsinsen_down, lambda : to_base64(base.pjoin('tsinsen-oj', prob.route, 'down.zip'))))
			for key, val in prob.tsinsen_files.items():
				f.write(add_shell('FileName', lambda : key.replace('/', '-').replace('\\', '-')))
				f.write(add_shell('File(%s)' % val, lambda : to_base64(base.pjoin(prob.path, 'resources', key))))
		if base.start_file:
			base.xopen_file(result_file)

def tuoj_down(conf = None):
	if not conf:
		conf = base.conf
		if not os.path.exists('tuoj'):
			base.remkdir('tuoj')
		base.remkdir(pjoin('tuoj', 'down'))
	if conf.folder == 'contest':
		for day in base.days():
			os.makedirs(base.pjoin(pjoin('tuoj', 'down'), day['name']))
			tuoj_down(day)
		return
	for prob in conf.probs():
		log.info(prob.route)
		os.makedirs(base.pjoin(pjoin('tuoj', 'down'), prob.route))
		for idx, case in enumerate(prob.sample_cases, start = 1):
			shutil.copy(
				base.pjoin(prob.path, 'down', str(case) + '.in'),
				base.pjoin('tuoj', 'down', prob.route, str(case) + '.in')
			)
			shutil.copy(
				base.pjoin(prob.path, 'down',str(case) + '.ans'),
				base.pjoin('tuoj', 'down', prob.route, str(case) + '.ans')
			)
		base.run_r(base.dos2unix, base.pjoin(pjoin('tuoj', 'down', prob.route)))

def loj_prob(conf, pre = False):
	if not pre and len(conf.get('pre', [])) > 0:
		loj_prob(conf, True)
	global save_flag
	headers = {
		'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36'
	}
	tool_conf = base.tool_conf[base.work]['default']
	cookies = tool_conf['cookies']
	pid_key = 'pid' if not pre else 'pid-pre'
	tag_key = 'tags' if not pre else 'tags-pre'
	pid = conf.get(pid_key, {}).get(base.work + '-default', 0)
	host = tool_conf['main']
	data = {
		'file_name' : conf['name'],
		'title' : conf.tr('title') + (u'（预测试）' if pre else ''),
		'description' : 'place holder',
		'input_format' : 'place holder',
		'output_format' : 'place holder',
		'example' : 'place holder',
		'limit_and_hint' : 'place holder',
		'tags' : conf.get(tag_key, {}).get(base.work + '-default', [])
	}
	def file_callback(m):
		p = (m.bytes_read / m.len) * 100
		print('上传 %.1f%%（%.1f KiB/%.1f KiB）    ' % (p, m.bytes_read / 1024, m.len / 1024), end = '\r')
		sys.stdout.flush()
	def post(url, data = None, files = None):
		if files:
			d = dict(files)
			if not data:
				data = {}
			for k, v in data.items():
				d[k] = str(v)
			d = MultipartEncoder(d)
			d = MultipartEncoderMonitor(d, file_callback)
			h = headers.copy()
			h['Content-Type'] = d.content_type
		else:
			d = data
			h = headers
		r = requests.post(
			host + url,
			headers = h,
			stream = True,
			cookies = cookies,
			data = d
		)
		if not r.ok:
			log.error(u'网站连接失败，错误代码%s，错误信息见`error.log`。' % r.status_code)
			with open('error.log', 'a') as f:
				f.write(u'运行时间：' + str(datetime.datetime.now()) + '\n')
				f.write(u'url：' + r.url + '\n')
				f.write(r.text)
		return r
	if pid == 0:
		r = post('/problem/%d/edit' % pid, data)
		pid = int(r.url.split('/')[-1])
		conf.setdefault(pid_key, {})
		conf[pid_key][base.work + '-default'] = pid
		save_flag = True
	path = pjoin('statements', base.work, conf.route if conf.route != '' else conf['name']) + '.md'
	data['description'] = open(path, 'rb').read()
	post('/problem/%d/edit' % pid, data)
	import zipfile, uuid
	conf_data = conf.data if not pre else conf.pre
	packed = conf.packed if not pre else conf.packed_pre
	cases = conf.test_cases if not pre else conf.pre_cases
	data_yml = {
		'inputFile' : '#.in',
		'outputFile' : '#.ans',
		'subtasks' : [{
			'score' : datum.score if packed else 100 / len(cases) * len(datum['cases']),
			'type' : 'min' if packed else 'sum',
			'cases' : [str(c) for c in datum['cases']]
		} for datum in conf_data]
	}
	if os.path.exists(pjoin(conf.path, 'data', 'chk', 'chk.cpp')):
		data_yml['specialJudge'] = {
			'language' : 'cpp17',
			'fileName' : 'spj_cpp.cpp'
		}
	if conf['type'] == 'output':
		data_yml['userOutput'] = (conf['name'] if conf.get('file io') else '') + '#.out'
	if conf.get('extra'):
		files = [{'name' : exname, 'dest' : exname} for exname in conf['extra']]
		data_yml['extraSourceFiles'] = [
			{'language' : lan, 'files' : files} \
			for lan in ['cpp', 'c', 'cpp11', 'cpp17', 'cpp14']
		]
	data_path = 'data' if not pre else 'pre'
	open(pjoin(base.work, data_path, conf.route + '.yml'), 'wb').write(base.dump_formats['yaml'](data_yml))
	def pack(z, path, fname, force_file = False):
		full_path = pjoin(path, fname)
		print(u"打包文件`%s`  " % full_path, end = '\r')
		sys.stdout.flush()
		if os.path.isdir(full_path):
			if not force_file:
				for sub in os.listdir(full_path):
					pack(z, path, pjoin(fname, sub))
				return
			else:
				log.warning(u'支持版本的`%s`不能是文件夹，用空文件代替。' % pjoin)
				id = str(uuid.uuid4()) + '.tmp'
				open(id, 'w')
		else:
			id = str(uuid.uuid4()) + '.tmp'
			base.shutil_copy(full_path, id)
		time.sleep(0.1)
		base.dos2unix(id, full_path)
		z.write(id, fname)
		os.remove(id)
	data_path = 'data' if not pre else 'pre'
	with zipfile.ZipFile(pjoin(base.work, data_path, conf.route + '.zip'), 'w') as z:
		for id in (conf.test_cases if not pre else conf.pre_cases):
			pack(z, pjoin(conf.path, data_path), id + '.in', force_file = True)
			pack(z, pjoin(conf.path, data_path), id + '.ans', force_file = True)
		if os.path.exists(pjoin(conf.path, 'data', 'chk', 'chk.cpp')):
			z.write(pjoin(conf.path, 'data', 'chk', 'chk.cpp'), 'spj_cpp.cpp')
		if os.path.exists(pjoin(base.work, data_path, conf.route + '.yml')):
			z.write(pjoin(base.work, data_path, conf.route + '.yml'), 'data.yml')
		if conf.get('extra'):
			for exname in conf['extra']:
				z.write(pjoin(conf.path, 'extra', exname), exname)
	files = [
		('testdata', ("data.zip", open(pjoin(base.work, data_path, conf.route + '.zip'), "rb"))),
	]
	with zipfile.ZipFile(pjoin(base.work, 'resources', conf.route + '.zip'), 'w') as z:
		try:
			for name in os.listdir(pjoin(conf.path, 'resources')):
				pack(z, pjoin(conf.path, 'resources'), name)
		except Exception as e:
			log.info(u'没有找到资源文件。')
	if os.path.exists(pjoin(conf.path, 'down')) and len(list(os.listdir(pjoin(conf.path, 'down')))) > 0:
		with zipfile.ZipFile(pjoin(base.work, 'down', conf.route + '.zip'), 'w') as z:
			if conf['type'] == 'output':
				for id in conf.test_cases:
					pack(z, conf.path, pjoin('data', id + '.in'))
				for name in os.listdir(pjoin(conf.path, 'down')):
					pack(z, conf.path, pjoin('down', name))
			else:
				for name in os.listdir(pjoin(conf.path, 'down')):
					pack(z, pjoin(conf.path, 'down'), name)
		files.append(('additional_file', ("down.zip", open(pjoin(base.work, 'down', conf.route + '.zip'), "rb"))))
	data = {
		'type' : {'program' : 'traditional', 'output' : 'submit-answer', 'interactive' : 'interaction', 'hand' : 'hand'}.get(conf['type'], conf['type'])
	}
	if conf['type'] != 'output':
		data['time_limit'] = int(conf.get('time limit', 0) * 1000)
		data['memory_limit'] = int(conf.ml().MB)
	if conf.get('file io'):
		data['io_method'] = 'file-io'
		data['file_io_input_name'] = conf.get('input_table', {}).get('zh-cn', conf['name'] + '.in')
		data['file_io_output_name'] = conf.get('output_table', {}).get('zh-cn', conf['name'] + '.out')
	post('/problem/%d/manage' % pid, data, files)
	files = [
		('images', ("resources.zip", open(pjoin(base.work, 'resources', conf.route + '.zip'), "rb")))
	]
	post('/problem/%d/upload_resource' % pid, files = files)

def loj():
	global save_flag
	save_flag = False
	syz_host = base.tool_conf.get(base.work, {}).get('default')
	if not syz_host:
		log.error(u'没有配置%s的地址。' % base.work)
		if base.system == 'Windows':
			log.info(u'在tuack的安装目录中找到conf.json。')
		else:
			log.info(u'在~/.tuack中找到conf.json。')
		log.info(u'添加%s的相关字段，详见`https://git.thusaac.com/publish/tuack/wikis/导出题目`。' % base.work)
		return
	if not os.path.exists(base.work):
		base.remkdir(base.work)
	base.remkdir(pjoin(base.work, 'data'))
	base.remkdir(pjoin(base.work, 'pre'))
	base.remkdir(pjoin(base.work, 'down'))
	base.remkdir(pjoin(base.work, 'resources'))
	if base.conf.folder == 'contest':
		for day in base.days():
			os.makedirs(base.pjoin(pjoin(base.work, 'data'), day.route))
			os.makedirs(base.pjoin(pjoin(base.work, 'down'), day.route))
			os.makedirs(base.pjoin(pjoin(base.work, 'pre'), day.route))
			os.makedirs(base.pjoin(pjoin(base.work, 'resources'), day.route))
	if base.do_render:
		from . import ren
		tmp = base.start_file
		base.start_file = False
		ren.Markdown(base.work).run()
		base.start_file = tmp
	else:
		pass
	for prob in base.probs():
		try:
			loj_prob(prob)
		except Exception as e:
			traceback.print_exc()
			log.error(u'输出题目`%s`时发生错误`%s`，具体信息如上所示。' % (prob['name'], e))
	if save_flag:
		base.save_json(base.conf)

def domjudge():
	global save_flag
	save_flag = False
	base.remkdir(base.work)
	if base.conf.folder == 'contest':
		for day in base.days():
			os.makedirs(pjoin(base.work, day.route))
	for prob in base.probs():
		try:
			os.makedirs(pjoin(base.work, prob.route))
			os.makedirs(pjoin(base.work, prob.route, 'data'))
			os.makedirs(pjoin(base.work, prob.route, 'data', 'secret'))
			os.makedirs(pjoin(base.work, prob.route, 'data', 'sample'))
			for cases, s_folder, t_folder in [
				(prob.sample_cases, 'down', 'sample'),
				(prob.test_cases, 'data', 'secret')
			]:
				for case in cases:
					for suf in ('.in', '.ans'):
						s_name = pjoin(prob.path, s_folder, str(case) + suf)
						t_name = pjoin(base.work, prob.route, 'data', t_folder, str(case) + suf)
						base.shutil_copy(s_name, t_name)
				base.run_r(base.dos2unix, pjoin(base.work, prob.route, 'data', t_folder))
			open(pjoin(base.work, prob.route, 'problem.yaml'), 'wb').write(base.dump_formats['yaml']({
				'problem_format_version': '2023-07-draft',
				'name': {
					'en': prob.get('title', {}).get('en', prob.get('name', '')),
					# 'zh': Voorbeeld probleem
				},
				'uuid': str(uuid.uuid4()),
				'type': 'pass-fail',
				'author': 'HDU',
				'source': 'SCCPC 2025',
				'source_url': '',
				'license': '',
				'rights_owner': '',
				'limits': {
					'time_multipliers': {
						'ac_to_time_limit': 2.0,
						'time_limit_to_tle': 1.5
					},
					'time_limit': prob.get('time limit', 1.0),
					'time_resolution': 1.0,
					'memory': prob.memory_limit().MB,
					'output': 1024,
					'code': 128,
					'compilation_time': 60,
					'compilation_memory': 2048,
					'validation_time': 60,
					'validation_memory': 2048,
					'validation_output': 8
				},
				# 'validation': 'custom',
				'keywords': [],
				'languages': ['c', 'cpp', 'java', 'kotlin', 'python3']
			}))
		except Exception as e:
			traceback.print_exc()
			log.error(u'输出题目`%s`时发生错误`%s`，具体信息如上所示。' % (prob['name'], e))
	if save_flag:
		base.save_json(base.conf)

work_list = {
	'lemon' : lemon,
	'arbiter' : arbiter,
	'arbiter-main' : arbiter_main,
	'arbiter-down' : arbiter_down,
	'tsinsen-oj' : tsinsen_oj,
	'tuoj-down' : tuoj_down,
	'loj' : loj,
	'ipuoj' : loj,
	'domjudge' : domjudge
}

if __name__ == '__main__':
	try:
		if base.init():
			for base.work in base.works:
				base.run_exc(work_list[base.work])
		else:
			log.info(u'这个工具用于导出成其他类型的工程。')
			log.info(u'支持的工作：%s' % ','.join(sorted(work_list.keys())))
	except base.NoFileException as e:
		log.error(e)
		log.info(u'尝试使用`python -m tuack.gen -h`获取如何生成一个工程。')



# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import re
import sys
import json
import datetime
import shutil
import subprocess
import time
import signal
import zipfile
from multiprocessing import Process, Queue
from functools import wraps
from threading import Timer
import platform
import logging
import traceback
import yaml
from inspect import isfunction, getsource

python_version = (sys.version_info.major, sys.version_info.minor, sys.version_info.micro)
if python_version[0] == 2:
	reload(sys)
	sys.setdefaultencoding('utf-8')

class Memory(str):
	'''
	'5KB'，'10 MB'，'8M'格式的字符串s，允许用Memory(s).GB，Memory(s).B等形式转换单位
	'''
	units = {
		'B' : 1,
		'K' : 2 ** 10,
		'KB' : 2 ** 10,
		'Ki' : 2 ** 10,
		'KiB' : 2 ** 10,
		'M' : 2 ** 20,
		'MB' : 2 ** 20,
		'Mi' : 2 ** 20,
		'MiB' : 2 ** 20,
		'G' : 2 ** 30,
		'GB' : 2 ** 30,
		'Gi' : 2 ** 30,
		'GiB' : 2 ** 30,
		'T' : 2 ** 40,
		'TB' : 2 ** 40,
		'Ti' : 2 ** 40,
		'TiB' : 2 ** 40,
		'P' : 2 ** 50,
		'PB' : 2 ** 50,
		'Pi' : 2 ** 50,
		'PiB' : 2 ** 50
	}
	def __new__(self, val):
		return super(Memory, self).__new__(self, val)
	def byte(self):
		if self[-1] == 'B':
			sp = 2 if self[-2] in self.units else 3 if self[-3:-1] in self.units else 1
		else:
			sp = 1
		un = self.units[self[-sp:]]
		return float(self[:-sp]) * un
	def __init__(self, val):
		super(Memory, self).__init__()
		b = self.byte()
		for key, val in self.units.items():
			self.__setattr__(key, b / val)

json_version = 2
work = None
system = platform.system()
out_system = system
def get_linux_version():
	try:
		return open('/etc/issue').read().split('')[0]
	except:
		return 'Unknown'
if system == 'Linux':
	linux_version = get_linux_version()
windows_stack_size = 536870912
diff_tool = 'diff' if system != 'Windows' else 'fc'
time_multiplier = 3.
elf_suffix = '' if system != 'Windows' else '.exe'
problem_skip = re.compile(r'^(data|down|tables|resources|gen|pre)$')
user_skip = re.compile(r'^(data|down|pre|val|.*validate.*|gen|chk|checker|report|check.*|make_data|data_maker|data_make|make|dmk|generate|generator|makedata|spj|judge|tables|tmp|cp|copy|mv|move|rm|remove|.*\.tmp|.*\.temp|temp|.*\.test|.*\.dir)(\..*)?$')
compilers = {
	'cpp' : lambda name, args, macros = '', ml = Memory('512 MB'): 'g++ %s.cpp -o %s %s %s %s' % (name, name, args, macros, '' if system != 'Windows' else '-Wl,--stack=%d' % int(ml.B)),
	'c' : lambda name, args, macros = '', ml = Memory('512 MB'): 'gcc %s.c -o %s %s %s %s' % (name, name, args, macros, '' if system != 'Windows' else '-Wl,--stack=%d' % int(ml.B)),
	# I don't know how to change stack size, nor add #define in pascal
	'pas' : lambda name, args, macros = '', ml = Memory('512 MB'): 'fpc %s.pas %s' % (name, args),
	'java' : lambda name, args, macros = '', ml = Memory('512 MB'): 'javac %s.java %s -J-Xms%dm -J-Xmx%dm' % (name, args, int(Memory(ml).MB) // 16, int(Memory(ml).MB) // 4),
	'py' : lambda name, args, macros = '', ml = Memory('512 MB'): ''
}

# 必须用函数，而不是lambda表达式才能并行
def java_runner(name, ml = None):
	return 'java -Xms%dm -Xmx%dm %s' % (int(Memory(ml).MB) // 16, int(Memory(ml).MB), name)

def py_runner(name, ml = None):
	return 'python3 %s.py' % name

runners = {
	'cpp' : None,
	'c' : None,
	'pas' : None,
	'java' : java_runner,
	'py' : py_runner
}
macros = {
	'uoj' : '-DONLINE_JUDGE',
	'noi' : '-D__ARBITER__',
	'release' : '',
	'test' : '-D__TUACK__ -DONLINE_JUDGE'
}

copied_data = set()
no_compiling = False
path = os.path.dirname(os.path.realpath(__file__))

log = logging.getLogger()
log.setLevel(logging.DEBUG)

class NoFileException(Exception):
	pass

pjoin = lambda *args : os.path.join(*args).rstrip('/').rstrip('\\')
rjoin = lambda *args : '/'.join(args).strip('/')

natsort_warned = None

home_path = os.path.expanduser("~")
try:
	tool_path = (path if system == 'Windows' and not home_path else pjoin(home_path, '.tuack'))
except:
	tool_path = path
if not os.path.exists(tool_path):
	os.makedirs(tool_path)

if system == 'Windows':
	format_checker_name = 'format-win.exe'
elif system == 'Darwin':
	format_checker_name = 'format-mac'
else:
	format_checker_name = 'format-linux'

if system == 'Windows':
	if home_path:
		tmp_path = pjoin(home_path, '.tuack_tmp')
	else:
		tmp_path = pjoin(os.getcwd(), 'tmp')
else:
	tmp_path = '/tmp/tuack'
if not os.path.exists(tmp_path):
	os.makedirs(tmp_path)

class Configure(dict):
	'''
	描述一个conf.json的对象
	'''
	@staticmethod
	def merge_item(base, ext):
		'''
		用于合并继承和原始的dict元素
		'''
		if type(base) == type(ext) == list:
			return base + ext
		elif type(base) != dict or type(ext) != dict:
			log.error('Extend conf.json error, type of key `%s` doesn\'t match.' % key)
			raise TypeError('extend error %s' % key)
		ret = base.copy()
		for key, val in ext.items():
			if key.endswith('+'):
				k = key[:-1]
				ret[k] = merge_item(base[k], val)
			else:
				ret[key] = val
		return ret

	def __init__(self, val, path = None, parent = None):
		if type(val) == dict:
			super(Configure, self).__init__(val)
		elif type(val) == str:
			super(Configure, self).__init__(json.loads(val))
		elif type(val) == bytes:
			super(Configure, self).__init__(json.loads(val.decode('utf-8')))
		else:
			raise Exception('Can\'t translate this object to a Configure.')
		self.folder = self['folder']
		self.parent = parent
		self.language = None
		self.path = path
		self.sub = []

	def lang(self):
		return self.language

	def tr(self, key):
		val = self[key]
		if type(val) != dict:
			return val
		if lang and lang in val:
			return val[lang]
		if self.lang() and self.lang() in val:
			return val[self.lang()]
		for k, v in val.items():
			return v
		return None

	def __contains__(self, key):
		return super(Configure, self).__contains__(key + '+') or super(Configure, self).__contains__(key) or (self.parent and key in self.parent)

	def getitem(self, key, trans = lambda prob, val, key, depth : val, depth = 0):
		if super(Configure, self).__contains__(key + '+'):
			cur = trans(self, super(Configure, self).__getitem__(key + '+'), key, depth)
			return self.merge_item(cur, self.parent.getitem(key, trans, depth + 1)) if self.parent else cur
		if super(Configure, self).__contains__(key):
			return trans(self, super(Configure, self).__getitem__(key), key, depth)
		return self.parent.getitem(key, trans, depth + 1) if self.parent else None

	def __getitem__(self, key):
		return self.getitem(key)

	def set_default(self, path = None):
		self.all = True
		if 'name' not in self:
			self['name'] = path
		return self

	def probs(self, pick = False, no_repeat = False):
		if no_repeat == True:
			no_repeat = set()
		if no_repeat != False:
			path = os.path.abspath(self.path)
			if path in no_repeat:
				return
			no_repeat.add(path)
		pick |= not sub_set or self.route in sub_set
		if type(self) == Problem:
			if pick or any_prefix(self.route):
				self.all = pick
				yield self
		else:
			for sub in self.sub:
				for prob in sub.probs(pick, no_repeat = no_repeat):
					yield prob

	def days(self, pick = False, no_repeat = False):
		if no_repeat == True:
			no_repeat = set()
		if no_repeat != False:
			path = os.path.abspath(self.path)
			if path in no_repeat:
				return
			no_repeat.add(path)
		pick |= not sub_set or self.route in sub_set
		if type(self) == Problem:
			return
		if type(self) == Day:
			if pick or any_prefix(self.route):
				self.all = pick
				yield self
		elif type(self) == Contest:
			for sub in self.sub:
				if type(sub) == Day:
					for day in sub.days(pick, no_repeat = no_repeat):
						yield day

	def name_lang(self):
		return self['name'] + ('-' + globals()['lang'] if globals()['lang'] else '')

def probs(item = None, pick = False, no_repeat = False):
	if not item:
		item = conf
	return conf.probs(pick, no_repeat)

def days(item = None, pick = False, no_repeat = False):
	if not item:
		item = conf
	return conf.days(pick, no_repeat)

class Contest(Configure):
	def __init__(self, *args):
		super(Contest, self).__init__(*args)

class Day(Configure):
	def __init__(self, *args):
		super(Day, self).__init__(*args)

class DataPath(str):
	def __new__(self, val):
		return super(DataPath, self).__new__(self, '*' * val['depth'] + val['case'])
	def __init__(self, val):
		self.data = val
		super(DataPath, self).__init__()
	def __getitem__(self, key):
		return self.data[key]
	def full(self):
		return pjoin(self['prefix'], self['key'], self['case'])

def sorter():
	global natsort_warned
	try:
		if not natsort_warned:
			check_install('natsort')
		import natsort
		return lambda inp : natsort.natsorted(inp, alg = natsort.ns.IGNORECASE)
	except:
		if not natsort_warned:
			log.warning(u'`natsort`用于给测试点名称排序，不使用的话可能会出现10排在2前面的情况。')
			natsort_warned = True
		return sorted

class Datum(dict):
	def __init__(self, datum, prob = None, key = 'data'):
		super(Datum, self).__init__(datum)
		self.prob = prob
		self.key = key
	def ml(self):
		return Memory(self['memory limit']) if 'memory limit' in self else self.prob.ml()

class Problem(Configure):
	def statement(self, cur_lang = None):
		if not cur_lang and lang:
			cur_lang = lang
		if cur_lang:
			source = pjoin(self.path, 'statement', cur_lang + '.md')
			if os.path.exists(source):
				self.language = cur_lang
				return source
		for source, self.language in (
			(pjoin(self.path, 'statement', 'zh-cn.md'), 'zh-cn'),
			(pjoin(self.path, 'statement', 'en.md'), 'en'),
			(pjoin(self.path, 'description.md'), None)
		):
			if os.path.exists(source):
				return source
		raise NoFileException('No md file found.')
	def lang(self):
		try:
			self.statement()
			return self.language
		except NoFileException as e:
			return None
	def __init__(self, *args):
		super(Problem, self).__init__(*args)
		self.chk = None
		self.data = self.get_data('data')
		self.samples = self.get_data('samples')
		self.pre = self.get_data('pre')
	def set_default(self, path = None):
		super(Problem, self).set_default(path)
		if 'title' not in self and 'cnname' in self:
			self['title'] = {'zh-cn' : self.pop('cnname')}
		for data, cases, attr, key in [('data', 'test cases', 'test_cases', 'data'), ('samples', 'sample count', 'sample_cases', 'down'), ('pre', 'pre count', 'pre_cases', 'pre')]:
			tc = set()
			if hasattr(self, data) and self.__getattribute__(data) != None:
				for datum in self.__getattribute__(data):
					tc |= set(datum['cases'])
			else:
				self[data] = []
			if cases in self and type(self[cases]) == int:
				log.warning(u'`%s`字段不再使用，使用`python -m tuack.gen upgrade`升级。' % cases)
				#to_dp = lambda i : DataPath({'case' : self['name'] + str(i), 'key' : key, 'depth' : 0, 'prefix' : self.path})
				#self[data] = [{'cases' : [
				#	to_dp(i) for i in range(1, self.pop(cases) + 1) \
				#	if to_dp(i) not in tc
				#]}]
			self.__setattr__(attr, sorter()(list(tc)))
	def set_score(self):
		if 'packed' in self:
			log.warning(u'题目`%s`的`packed`字段不再有效，但仍可以存在，用`python -m tuack.gen upgrade`升级。' % self.route)
		for key, packed in [('data', ''), ('pre', '_pre'), ('down', '_sample')]:
			if not hasattr(self, key):
				self.__setattr__('packed' + packed, False)
				return
			num_unscored = 0
			total_score = 0.0
			data_objs = self.__getattribute__(key)
			if data_objs is None:
				continue
			for datum in data_objs:
				if 'score' in datum:
					datum.score = datum['score']
					total_score += datum['score']
				else:
					num_unscored += 1
			if num_unscored != 0:
				item_score = (100. - total_score) / num_unscored
				for datum in data_objs:
					if 'score' not in datum:
						datum.score = item_score
			if num_unscored == len(data_objs):
				self.__setattr__('packed' + packed, False)
				self.__setattr__('score' + packed, 100.)
			else:
				self.__setattr__('packed' + packed, True)
				if num_unscored != 0:
					log.warning(u'题目`%s`有%d个包设置了`score`，有%d个没有；总分共设%f分，将剩下%f分平分给没有的包。' % (
						self.route,
						len(data_objs) - num_unscored,
						num_unscored,
						total_score,
						100 - total_score
					))
					log.info(u'如果你需要不等分+打包测试，请每个包设置`score`；否则请每个包都不设置`score`，此时是每个测试点同分而不是每个包同分。')
					log.info(u'一部分包设置`score`，另一部分不设置将可能导致导出其他格式时出现问题。')
					if key != 'data':
						log.warning(u'以上信息属于这道题的%s，不属于标准测试点。' % {'pre' : u'预测试点', 'down': u'样例'}[key])
					self.__setattr__('score' + packed, 100.)
				else:
					if abs(total_score - 100) > 1e-6:
						log.warning(u'题目`%s`总分是%f分，不是100分。' % (self.route, total_score))
						if key != 'data':
							log.warning(u'以上信息属于这道题的%s，不属于标准测试点。' % {'pre' : u'预测试点', 'down': u'样例'}[key])
					self.__setattr__('score' + packed, total_score)

	def extend_pathed(self, path):
		if path.startswith(':'):
			return self.parent.extend_pathed(path[1:])
		return pjoin(self.path, path)

	def users(self):
		def users_pathed(self, users, key = '', depth = 0):
			if type(users) != dict:
				log.warning(u'`json.conf`中，`users`项%s:%s已经过时，但仍可以使用，用`python -m tuack.gen upgrade`升级' % (key, users))
				return {
					'path' : self.extend_pathed(users),
					'expected' : {}
				}
			elif depth == 2:
				# Is detailed description
				if "path" not in users:
					# Ill-formed
					log.warning(u'`users`项%s:%s不含`path`项' % (key, users))
					return None
				return {
					'path' : self.extend_pathed(users["path"]),
					'expected' : users["expected"] if "expected" in users else {}
				}

			return {
				key : val
				for key, val in \
				((key, users_pathed(self, val, key, depth + 1)) for key, val in users.items()) \
				if val
			}
		return self.getitem('users', users_pathed)
	
	def expect(self, user, algo, score):
		'''
		Check expected scores
		'''
		exp = self.users()[user][algo]['expected']
		algo_failed = False
		if type(exp) == dict:
			for symbol, value in exp.items():
				algo_failed |= not eval('round(score, 8) %s %s' % (symbol, value))
		elif type(exp) == list:
			for pred in exp:
				algo_failed |= not eval('round(score, 8) %s' % pred)
		elif type(exp) == str:
			algo_failed |= not eval('round(score, 8) %s' % exp)
		elif exp is None:
			pass
		else:
			log.warning('`expected`字段必须是字符串、数组或字典。')
		return not algo_failed

	def get_data(self, key):
		def data_pathed(self, data, key, depth):
			key_map = {
				'data' : 'data',
				'samples' : 'down',
				'pre' : 'pre'
			}
			ret = []
			for datum in data:
				tmp = []
				for case in datum['cases']:
					cur = self
					dep = depth
					cas = str(case)
					while cas.startswith(':'):
						cas = cas[1:]
						dep -= 1
						cur = cur.parent
					tmp.append(DataPath({'case' : cas, 'key' : key_map[key], 'depth' : dep, 'prefix' : cur.path}))
				datum_fixed = datum.copy()
				datum_fixed['cases'] = tmp
				ret.append(Datum(datum_fixed, self, key))
			return ret
		return self.getitem(key, data_pathed)

	def ml(self):
		return Memory(self.get('memory limit', '2 GB'))

	def memory_limit(self):
		return self.ml()

#mem_json = {}	# 题目复用的话，需要记忆化

def load_json(path = '.', route = None):
	#abs_path = os.path.abspath(path)
	#if abs_path in mem_json:
	#	return mem_json[abs_path]
	for name in ['conf.yaml', 'conf.json', 'prob.json']:
		try:
			full_path = pjoin(path, name)
			if os.path.isfile(full_path):
				inp = open(full_path, 'rb').read().decode('utf-8')
				if name.endswith('json'):
					conf = json.loads(inp)
				elif name.endswith('yaml'):
					conf = yaml.full_load(inp)
				if conf.get('version', -1) > json_version:
					log.warning(u'`conf.*`版本高于`tuack`，请升级`tuack`。')
				if 'folder' not in conf:
					conf['folder'] = 'problem'
				args = [conf, path]
				if conf['folder'] == 'extend':
					base_conf = load_json(pjoin(path, conf['base path']))
					folder = base_conf.folder
					args.append(base_conf)
				else:
					folder = conf['folder']
				conf = eval(folder.capitalize())(*args)
				if conf['folder'] == 'extend':
					conf.folder = base_conf.folder
				conf.set_default(os.path.basename(path))
				conf.route = '' if route == None else rjoin(route, conf['name'])
				conf.full_path = os.path.abspath(conf.route)
				if conf.folder == 'problem':
					conf.set_score()
				if 'subdir' in conf:
					conf.sub = []
					for sub in conf['subdir']:
						ret = load_json(pjoin(path, sub), conf.route)
						if ret:
							conf.sub.append(ret)
				#mem_json[abs_path] = conf
				return conf
		except Exception as e:
			log.warning(u'文件`%s`错误。' % pjoin(path, name))
			log.info(e)
			log.info('可能会导致运行到该目录下出现问题。')
	else:
		#raise NoFileException(u'路径`%s`下找不到conf.*。' % path)
		log.warning(u'路径`%s`下找不到conf.*。' % path)
		log.info('可能会导致运行到该目录下出现问题。')

def del_redundance(conf, red):
	for key in red:
		try:
			conf.pop(key)
		except:
			pass
	if not conf['name'] or conf['name'] == '.':
		conf.pop('name')
	return conf

def to_ori(x):
	if type(x) != dict and type(x) != list: return
	for key in (x if issubclass(type(x), dict) else range(len(x))):
		for tp in [int, float, str, dict, list]:
			if issubclass(type(x[key]), tp) and type(x[key]) != bool:
				x[key] = tp(x[key])
		to_ori(x[key])
	return x

dump_formats = {
	'json' : lambda conf : json.dumps(conf, indent = 2, sort_keys = True, ensure_ascii = False).encode('utf-8'),
	'yaml' : lambda conf : yaml.safe_dump(to_ori(dict(conf)), encoding = 'utf-8', allow_unicode = True)
}

def save_json(conf):
	for s in conf.sub:
		save_json(s)
	try:
		open(pjoin(conf.path, 'conf.' + dump_format), 'wb').write(dump_formats[dump_format](conf))
	except Exception as e:
		log.error(u'不支持配置文件格式`%s`。' % dump_format)
		log.info(e)

def any_prefix(pre, st = None):
	if not st:
		st = sub_set
	for s in st:
		if s == pre:
			return 1
		if pre == '' or s.startswith(pre + '/'):
			return 2
	return 0

def mkdir(name):
	if not os.path.exists(name):
		os.makedirs(name)

def copy(source, name, target):
	full_source = pjoin(source, name)
	if not os.path.exists(full_source):
		raise NoFileException('No such file or path `%s`.' % full_source)
	copied_data.add(full_source)
	if os.path.isdir(full_source):
		if full_source.endswith('.dir') or no_compiling:
			full_target = (pjoin(target, name) if os.path.exists(target) else target)
			if full_source.endswith('.dir'):
				full_target = full_target[:-4]
			shutil.copytree(full_source, full_target)
		else:
			# TODO: make if there is a makefile
			ret = os.getcwd()
			os.chdir(pjoin(source, name))
			cpp_file = name + '.cpp'
			elf_file = name + elf_suffix
			check_install('g++')
			if os.system('g++ %s -o %s -O2 -std=c++14 -Wall' % (cpp_file, elf_file)) != 0:
				os.chdir(ret)
				log.error(u'`%s`编译失败。' % pjoin(full_source, cpp_file))
				return True
			else:
				os.chdir(ret)
				log.info(u'`%s`编译成功。' % pjoin(full_source, cpp_file))
			shutil.move(pjoin(full_source, elf_file), target)
	else:
		shutil.copy(full_source, target)
	return True

def shutil_copy(source, target):
	func = shutil.copytree if os.path.isdir(source) else shutil.copy
	func(source, target)

def xopen_file(path):
	try:
		if system == 'Windows':
			os.startfile(path)
		elif system == 'Darwin':
			subprocess.call(["open", path])
		else:
			subprocess.call(["xdg-open", path])
	except Exception as e:
		log.warning(u'打开文件`%s`失败。' % path)
		log.info(e)

def deal_args():
	global do_copy_files, do_test_progs, do_release, works, start_file, do_pack, langs, lang, sub_set, out_system, args, do_zip, do_encript, do_render, time_multiplier, git_lfs, dump_format, user_time, water_mark, noi_pas_c
	do_render = True
	works = []
	args = []
	langs = [None]
	lang = None
	sub_set = None
	start_file = True
	do_pack = True
	do_zip = False
	do_encript = False
	git_lfs = False
	user_time = False
	dump_format = 'yaml'
	water_mark = False
	noi_pas_c = False
	l = len(sys.argv)
	i = 1
	while i < l:
		if sys.argv[i] == '-i':
			i += 1
			os.chdir(sys.argv[i])
		elif sys.argv[i] == '-p':
			i += 1
			sub_set = set(sys.argv[i].split(','))
		elif sys.argv[i] == '-o':
			i += 1
			out_system = sys.argv[i]
		elif sys.argv[i] == '-d':
			i += 1
			dump_format = sys.argv[i]
		elif sys.argv[i] == '-s':
			start_file = False
		elif sys.argv[i] == '-k':
			do_pack = False
		elif sys.argv[i] == '-z':
			do_zip = True
		elif sys.argv[i] == '-g':
			git_lfs = True
		elif sys.argv[i] == '-r':
			do_render = False
		elif sys.argv[i] == '-e':
			do_zip = True
			do_encript = True
		elif sys.argv[i] == '-u':
			user_time = True
		elif sys.argv[i] == '-l':
			i += 1
			langs = set(sys.argv[i].split(','))
		elif sys.argv[i] == '-t':
			i += 1
			time_multiplier = float(sys.argv[i])
		elif sys.argv[i] == '--noi-pas-c':
			noi_pas_c = True
		elif sys.argv[i] == '-w':
			i += 1
			water_mark = sys.argv[i]
		elif sys.argv[i] == '-h' or sys.argv[i] == '--help':
			log.info(u'详细用法见文档：https://git.thusaac.com/publish/tuack/wikis。')
			log.info(u'python 脚本 [[[工作1],工作2],...] [[[选项1] 选项2] ...] [[[参数1] 参数2] ...]')
			log.info(u'工作必须在参数前面，工作用逗号隔开，选项和参数用空格隔开。')
			log.info(u'只有有逗号的项目可以用逗号获得多个结果，逗号前后不能有空白符。')
			log.info(u'这套工具的大多数脚本都可以在比赛、比赛日和题目目录下运行。')
			log.info(u'选项：')
			log.info(u'  -i PATH             指定PATH作为工作路径，否则使用当前路径。')
			log.info(u'  -s                  对于有输出的文件的操作，输出完以后不自动打开文件。')
			log.info(u'  -p day0/sleep,day2  只对day0/sleep和day2进行本次操作；此路径是基于当前文件夹的，')
			log.info(u'                      例如：在比赛日目录如day1下，则可以直接指定题目如exam；')
			log.info(u'                      对于test，还可以指定用户或算法，如day1/problem/vfk/std.cpp。')
			log.info(u'  -t 6.0              对于test，设置掐断时间为6.0*时间限制，用于对比不同程序的时限。')
			log.info(u'  -o SYSTEM           对于ren，输出指定操作系统的题面，可选Windows和Linux。')
			log.info(u'  -l zh-cn,en         对于ren，指定输出语言，不指定默认为zh-cn。')
			log.info(u'  -w logo.png         对于ren，给PDF添加水印，默认不添加。')
			log.info(u'  --noi-pas-c         对于ren渲染noi格式，强制在封面添加pas和c。')
			log.info(u'  -r                  对于dump，不先尝试渲染题面。')
			log.info(u'  -g                  对于gen，使用git-lfs。')
			log.info(u'  -d json             对于gen，规定配置文件格式，支持json、yaml，默认yaml。')
			log.info(u'  -u                  对于test，在linux下使用user time做测试，默认real time。')
			
			return False
		else:
			if len(works) == 0:
				works = sys.argv[i].split(',')
			else:
				args += sys.argv[i].split(',')
		i += 1
	return True

def custom_conf():
	get_tool_conf()
	c = env_conf['file_log'] if 'file_log' in env_conf else {}
	file_log = logging.FileHandler(
		c['path'] if 'path' in c else 'tuack.log',
		mode = c['mode'] if 'mode' in c else 'a',
		encoding = c['encoding'] if 'encoding' in c else None
	)
	file_log.setLevel(c['level'] if 'level' in c else logging.INFO)
	file_log.setFormatter(logging.Formatter(
		c['format'] if 'format' in c else '[%(levelname).1s]%(filename)s:%(funcName)s:%(lineno)d:%(message)s'
	))
	log.addHandler(file_log)

	c = env_conf['bash_log'] if 'bash_log' in env_conf else {}
	if 'encoding' in c:
		class MyStream(object):
			def __init__(self, stream):
				self.stream = stream
			def write(self, s):
				self.stream.buffer.write(s.encode(c['encoding']))
				self.stream.flush()
			def flush(self):
				self.stream.flush()
		bash_log = logging.StreamHandler(MyStream(sys.stdout))
	else:
		bash_log = logging.StreamHandler()
	bash_log.setLevel(c['level'] if 'level' in c else logging.DEBUG)
	bash_log.setFormatter(logging.Formatter(
		c['format'] if 'format' in c else '[%(levelname).1s]%(message)s'
	))
	log.addHandler(bash_log)

def init():
	import __main__
	global conf
	custom_conf()
	if not deal_args():
		return False
	log.info(
		('脚本%s，工程路径%s，参数%s，开始于%s。' if python_version[0] == 2 else u'脚本%s，工程路径%s，参数%s，开始于%s。') % (
			__main__.__file__, os.getcwd(), str(sys.argv[1:]), str(datetime.datetime.now())
		)
	)
	conf = load_json()
	try:
		check_install('git')
		check_install('git_lfs')
	except:
		pass
	return True

def tr(item):
	if 'zh-cn' in item:
		return item['zh-cn']
	elif 'en' in item:
		return item['en']
	elif len(item) > 1:
		for val in item.values():
			return val
	else:
		return 'Unknown'

def run_r(cmd, path):
	for f in os.listdir(path):
		cpath = pjoin(path, f)
		if os.path.isfile(cpath):
			cmd(cpath)
		else:
			run_r(cmd, cpath)

def get_tool_conf():
	def get_sys_env():
		return '$'.join([
			os.environ[key]
			for key in ['OS', 'SESSIONNAME', 'USERNAME', 'COMPUTERNAME', 'USERDOMAIN', 'USER', 'SHELL', 'SESSION'] \
			if key in os.environ
		] + ['py%x' % sys.hexversion])
	global env_conf
	try:
		tool_conf = json.loads(open(pjoin(tool_path, 'conf.json'), 'rb').read().decode('utf-8'))
	except:
		tool_conf = {}
	sys_env = get_sys_env()
	if sys_env not in tool_conf:
		tool_conf[sys_env] = {}
	env_conf = tool_conf[sys_env]
	return tool_conf

class NoInstalledException(Exception):
	def __init__(self, value):
		self.value = value
	def __str__(self):
		return repr(self.value)

def check_install(pack):
	def check_import(pack, extra_info = '', pack_name = None, level = logging.ERROR):
		try:
			__import__(pack)
		except Exception as e:
			log.log(level, u'python包%s没有安装，使用 pip install %s 安装。%s' % (pack, pack_name if pack_name else pack, extra_info))
			if system == 'Windows':
				log.info(u'如果pip没有安装，Windows下推荐用Anaconda等集成环境。')
			if system == 'Linux':
				log.info(u'如果pip没有安装，Ubuntu下用 sudo apt install python-pip 安装。')
			raise NoInstalledException(str(e))
	check_pyside = lambda : check_import('PySide', u'注意这个包只能在 python2 下使用。', 'pyside')
	check_natsort = lambda : check_import('natsort', level = logging.WARNING)
	def check_pandoc():
		ret = os.system('pandoc -v')
		if ret != 0:
			log.error(u'格式转换工具pandoc没有安装。')
			if system == 'Windows':
				log.info(u'Windows用户去官方网站下载安装，安装好后把pandoc.exe所在路径添加到环境变量PATH中。')
			if system == 'Linux':
				log.info(u'Ubuntu下用 sudo apt install pandoc 安装。')
			raise NoInstalledException('pandoc not found')
	def check_xelatex():
		ret = os.system('xelatex --version')
		if ret != 0:
			log.error(u'TeX渲染工具XeLaTeX没有安装。')
			if system == 'Windows':
				log.info(u'Windows下可以先安装MiKTeX，在首次运行的时候会再提示安装后续文件。')
			if system == 'Linux':
				log.info(u'Ubuntu下先用 sudo apt install texlive-xetex texlive-fonts-recommended texlive-latex-extra 安装工具；')
				log.info(u'然后一般会因为缺少有些字体而报错（Windows有使用权，但Ubuntu没有，所以没有预装这些字体）。')
				log.info(u'可以使用下列页面上的方法安装缺少的字体或是把win下的字体复制过来。')
				log.info(u'http://linux-wiki.cn/wiki/zh-hans/LaTeX%E4%B8%AD%E6%96%87%E6%8E%92%E7%89%88%EF%BC%88%E4%BD%BF%E7%94%A8XeTeX%EF%BC%89')
			raise NoInstalledException('xelatex not found')
	def check_git():
		ret = os.system('git --version')
		if ret != 0:
			log.warning(u'版本管理工具git没有安装，如果工程用git维护则你的修改可能无法成功提交。')
			log.info(u'一个可能的安装教程见这里：')
			log.info(u'https://git-scm.com/book/zh/v2/%E8%B5%B7%E6%AD%A5-%E5%AE%89%E8%A3%85-Git')
			if system == 'Windows':
				log.info(u'Windows下有多种不同的git版本，大家可以多交流好用的版本。')
			log.info(u'git入门可以参看这里：')
			log.info(u'http://rogerdudler.github.io/git-guide/index.zh.html')
			log.info(u'一般推荐用ssh方式克隆仓库，并用公私钥保证安全，添加密钥的方式一般仓库的git网页上能找到。')
			raise NoInstalledException('git not found')
	def check_git_lfs():
		ret = os.system('git lfs')
		if ret != 0:
			log.warning(u'git大文件系统lfs没有安装，如果工程使用了它你可能无法同步。')
			log.info(u'因为有些评测数据比较大，所以一般要求用git lfs大文件系统（Large File System）管理评测数据（in/ans）')
			log.info(u'一个可能的安装教程见这里：')
			log.info(u'https://git-lfs.github.com/')
			log.info(u'如果你用本工具的generator生成题目工程，那么你装好lfs以后一般可以不用再手工指定in和ans文件用lfs管理。')
			log.info(u'如果你的多人合作工程用到了lfs，请务必不要在没有安装lfs前把数据添加到工程中！')
			raise NoInstalledException('git lfs not found')

	def check_gpp():
		ret = os.system('g++ -v')
		if ret != 0:
			log.warning(u'g++未安装，将可能无法测试C++代码。')
			raise NoInstalledException('g++ not found')
	
	def check_gcc():
		ret = os.system('gcc -v')
		if ret != 0:
			log.warning(u'gcc未安装，将可能无法测试C代码。')
			raise NoInstalledException('gcc not found')
			
	def check_java():
		ret = os.system('javac -version')
		if ret != 0:
			log.warning(u'javac未安装，将可能无法测试Java代码。')
			raise NoInstalledException('javac not found')
		ret = os.system('java -version')
		if ret != 0:
			log.warning(u'java未安装，将可能无法测试Java代码。')
			raise NoInstalledException('java not found')

	def check_py2():
		ret = os.system('python2 --version')
		if ret != 0:
			log.warning(u'python2未安装，将可能无法测试Python2代码。')
			raise NoInstalledException('python2 not found')

	def check_py3():
		ret = os.system('python3 --version')
		if ret != 0:
			log.warning(u'python3未安装，将可能无法测试Python3代码。')
			raise NoInstalledException('python3 not found')

	def check_py():
		ret = os.system('python --version')
		if ret != 0:
			log.warning(u'python未安装，将可能无法测试Python代码。')
			raise NoInstalledException('python not found')
			
	def check_flex():
		ret = os.system('flex --version')
		if ret != 0:
			log.warning(u'flex未安装，将可能无法检查题面格式。')
			raise NoInstalledException('flex not found')

	def check_bison():
		ret = os.system('bison --version')
		if ret != 0:
			log.warning(u'bison未安装，将可能无法检查题面格式。')
			raise NoInstalledException('bison not found')
			
	def check_format():
		ret = os.system('"%s" -v' % pjoin(tool_path, format_checker_name))
		if ret != 0:
			log.warning(u'format checker未安装，使用`python -m tuack.install format`安装。')
			raise NoInstalledException('format not found')
			
	def check_unrar():
		ret = os.system('unrar --version')
		if ret != 0:
			log.warning(u'unrar未安装，将无法解压rar文件，请安装unrar。')
			log.warning(u'对于Windows用户，你可以将WinRar的安装目录添加到PATH。')
			raise NoInstalledException('unrar not found')

	def check_time():
		ret = os.system('time ls')
		if ret != 0:
			log.warning(u'time未安装，将无法测试运行时间。')
			log.warning(u'类Ubuntu用户可以用apt install time安装。')
			raise NoInstalledException('time not found')

	if pack in {'g++', 'cpp'}:
		pack = 'gpp'
	if pack == 'c':
		pack = 'gcc'
	global tool_conf
	tool_conf = get_tool_conf()
	if 'installed' not in env_conf:
		env_conf['installed'] = {}
	if pack in env_conf['installed'] and env_conf['installed'][pack]:
		return True
	try:
		eval('check_' + pack)
	except:
		check_import(pack)
	else:
		eval('check_' + pack)()
	env_conf['installed'][pack] = True

	open(pjoin(tool_path, 'conf.json'), 'wb').write(json.dumps(tool_conf, indent = 2, sort_keys = True).encode('utf-8'))
	return True

def change_eol(path, eol, show_path = None):
	import uuid
	ufname = str(uuid.uuid4()) + '.tmp'
	space_end = False
	is_text = True
	with open(ufname, 'wb') as f:
		try:
			for idx, line in enumerate(open(path, 'rb')):
				line = line.rstrip(b'\r\n')
				f.write(line + eol)
				if not space_end and (line.endswith(b' ') or line.endswith(b'\t')):
					log.warning(u'换行符转换：文件`%s`第%d行末尾有空白符。' % (show_path if show_path else path, idx + 1))
					space_end = True
				for code in ['utf-8', 'gbk']:
					try:
						if '\0' in line.decode(code):
							raise Exception('`\\0` in line')
						break
					except:
						pass
				else:
					is_text = False
					log.info(u'换行符转换：文件`%s`不是文本文件。' % (show_path if show_path else path))
					break
		except:
			is_text = False
	if is_text:
		os.remove(path)
		time.sleep(0.1)
		shutil.move(ufname, path)
	else:
		os.remove(ufname)

unix2dos = lambda path, show_path = None : change_eol(path, b'\r\n', show_path)
dos2unix = lambda path, show_path = None : change_eol(path, b'\n', show_path)

wiki = lambda name : name

def run_exc(func):
	try:
		return (True, func())
	except Exception as e:
		log.error(e)
		log.info(traceback.format_exc())
		return (False, None)

def rmtree(path, ignore_errors = True):
	if os.path.exists(path):
		if system == 'Windows':
			os.system(f'rmdir /s /q "{path}"')
		else:
			shutil.rmtree(path, ignore_errors = ignore_errors)

def repeat(func):
	las = None
	for i in range(10):
		try:
			func()
		except Exception as e:
			las = e
			time.sleep(1e-2)
			log.warning(f'执行该函数第{i + 1}次：' + func.__qualname__)
		else:
			return
	log.error('重复执行该函数多次失败：' + func.__qualname__)
	log.info(getsource(func))
	raise las

def remkdir(name):
	try:
		repeat(lambda : rmtree(name))
		os.makedirs(name)
	except Exception as e:
		log.warning('Can\'t delete %s' % name)
		log.warning(e)



# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import re
import sys
import json
import datetime
import shutil
import subprocess
import time
import signal
import zipfile
from multiprocessing import Process, Queue
from functools import wraps
from threading import Timer
import platform
from .base import *
from . import base
import math

doc_format = re.compile(r'^(.*)\.(doc|docs|ppt|pptx|pdf|tex|md|html|htm|zip|dir)$')
empty_cpp = 'int main(){}'

def find_doc(path, name):
	for f in os.listdir(path):
		m = doc_format.match(f)
		if m and m.group(1) == name:
			return name, '.' + m.group(2)
	return None

def test_copy_problem_files(prob):
	data_path = os.path.join(prob.path, 'data')
	try:
		copy(data_path, 'chk', base.pjoin(output_folder, prob.route))
		prob.chk = True
	except base.NoInstalledException as e:
		raise(e)
	except Exception as e:
		prob.chk = False
	
def copy_one_day_files(probs, day_name):
	remkdir(os.path.join(output_folder, day_name, 'data'))
	print('copy data files')
	for prob in probs:
		if base.rjoin(day_name, prob['name']) not in base.prob_set:
			continue
		# TODO: if test cases is a list of scores instead of an integer
		data_path = os.path.join(day_name, prob['name'], 'data')
		for i in range(1, prob['test cases'] + 1):
			for suf in ['.in', '.ans']:
				copy(data_path, prob['name'] + str(i) + suf, os.path.join(output_folder, day_name, 'data'))
		copy(data_path, 'chk', os.path.join(output_folder, day_name, 'data', prob['name'] + '_e'))
		for name in os.listdir(data_path):
			if os.path.join(data_path, name) not in base.copied_data:
				warning('Unusual file \'%s\' found.' % os.path.join(data_path, name))
				copy(data_path, name, os.path.join(output_folder, day_name, 'data'))
	print('dos2unix data files')
	if os.system('dos2unix %s 2> log' % os.path.join(output_folder, day_name, 'data', '*')) != 0:
		warning('dos2unix failed.')
	print('copy down files')
	remkdir(os.path.join(output_folder, day_name, 'down'))
	for prob in probs:
		if base.rjoin(day_name, prob['name']) not in base.prob_set:
			continue
		# TODO: if test cases is a list of scores instead of an integer
		data_path = os.path.join(day_name, prob['name'], 'down')
		target_path = os.path.join(output_folder, day_name, 'down', prob['name'])
		if not os.path.exists(target_path):
			os.makedirs(target_path)
		#print(prob)
		case_no = (prob['test cases'] if prob['type'] == 'output' else prob['sample count'])
		suffices = (['.in'] if prob['type'] == 'output' else ['.in', '.ans'])
		for i in range(1, case_no + 1):
			for suf in suffices:
				copy(data_path, prob['name'] + str(i) + suf, target_path)
		copy(data_path, 'checker', target_path)
		for name in os.listdir(data_path):
			if os.path.join(data_path, name) not in base.copied_data:
				warning('Unusual file \'%s\' found.' % os.path.join(data_path, name))
				copy(data_path, name, target_path)
	print('dos2unix down files')
	for prob in probs:
		if os.system('dos2unix %s 2> log' % os.path.join(output_folder, day_name, 'down', prob['name'], '*')) != 0:
			warning('dos2unix failed.')
			break
	remkdir(os.path.join(output_folder, day_name, 'discussion'))
	for prob in probs:
		res = find_doc(os.path.join(day_name, prob['name']), 'discussion')
		if res:
			copy(os.path.join(day_name, prob['name']), res[0] + res[1], os.path.join(output_folder, day_name, 'discussion', prob['name'] + res[1]))
		else:
			warning('Can\'t find discussion ppt for problem `%s`.' % prob['name'])
	if os.path.exists('log'):
		os.remove('log')
		
def pc2_copy_one_day_files(probs, day_name):
	remkdir(os.path.join('pc2', day_name))
	remkdir(os.path.join('pc2', day_name, 'data'))
	print('copy data files')
	for prob in probs:
		if base.rjoin(day_name, prob['name']) not in base.prob_set:
			continue
		name = prob['name'].capitalize()
		# TODO: if test cases is a list of scores instead of an integer
		remkdir(os.path.join('pc2', day_name, 'data', name))
		data_path = os.path.join(day_name, name, 'data')
		target_path = os.path.join('pc2', day_name, 'data', name)
		for i in range(1, prob['test cases'] + 1):
			for suf in ['.in', '.ans']:
				copy(data_path, prob['name'] + str(i) + suf, target_path)
		copy(data_path, 'chk', target_path)
		for name in os.listdir(data_path):
			if os.path.join(data_path, name) not in base.copied_data:
				warning('Unusual file \'%s\' found.' % os.path.join(data_path, name))
				copy(data_path, name, target_path)
		if os.system('dos2unix %s 2> log' % os.path.join(target_path, '*')) != 0:
			warning('dos2unix failed.')
	print('copy down files')
	remkdir(os.path.join('pc2', day_name, 'down'))
	for prob in probs:
		if base.rjoin(day_name, prob['name']) not in base.prob_set:
			continue
		name = prob['name'].capitalize()
		# TODO: if test cases is a list of scores instead of an integer
		data_path = os.path.join(day_name, prob['name'], 'down')
		target_path = os.path.join('pc2', day_name, 'down', name)
		remkdir(target_path)
		case_no = (prob['test cases'] if prob['type'] == 'output' else prob['sample count'])
		suffices = (['.in'] if prob['type'] == 'output' else ['.in', '.ans'])
		for i in range(1, case_no + 1):
			for suf in suffices:
				copy(data_path, prob['name'] + str(i) + suf, target_path)
		copy(data_path, 'checker', target_path)
		for name in os.listdir(data_path):
			if os.path.join(data_path, name) not in base.copied_data:
				warning('Unusual file \'%s\' found.' % os.path.join(data_path, name))
				copy(data_path, name, target_path)
	print('dos2unix down files')
	for prob in probs:
		if os.system('dos2unix %s 2> log' % os.path.join('pc2', day_name, 'down', prob['name'], '*')) != 0:
			warning('dos2unix failed.')
			break
	remkdir(os.path.join('pc2', day_name, 'discussion'))
	for prob in probs:
		res = find_doc(os.path.join(day_name, prob['name']), 'discussion')
		if res:
			copy(os.path.join(day_name, prob['name']), res[0] + res[1], os.path.join('pc2', day_name, 'discussion', prob['name'] + res[1]))
		else:
			warning('Can\'t find discussion ppt for problem `%s`.' % prob['name'])
	if os.path.exists('log'):
		os.remove('log')
		
def prob2uoj_conf(prob):
	s = ''
	s += 'use_builtin_judger on\n'
	if 'uoj checker' not in prob:
		s += 'use_builtin_checker wcmp\n'
	elif prob['uoj checker'] != False:
		s += 'use_builtin_checker %s\n' % prob['uoj checker']
	if prob['type'] == 'output':
		s += 'submit_answer on\n'
	s += 'n_tests %d\n' % prob['test cases']
	s += 'n_ex_tests %d\n' % prob['sample count']
	s += 'n_sample_tests %d\n' % prob['sample count']
	s += 'input_pre %s\n' % prob['name']
	s += 'input_suf in\n'
	s += 'output_pre %s\n' % prob['name']
	s += 'output_suf ans\n'
	if 'time limit' in prob:
		s += 'time_limit %d\n' % math.ceil(prob['time limit'])
	if 'memory limit' in prob:
		s += 'memory_limit %d\n' % (base.memory2bytes(prob['memory limit']) // 1024 // 1024)
	if 'output limit' in prob:
		s += 'output_limit %d\n' % (base.memory2bytes(prob['output limit']) // 1024 // 1024)
	else:
		s += 'output_limit 64\n'
	return s

def uoj_copy_one_day_files(probs, day_name):
	print('copy data files')
	for prob in probs:
		if base.rjoin(day_name, prob['name']) not in base.prob_set:
			continue
		# TODO: if test cases is a list of scores instead of an integer
		data_path = os.path.join(day_name, prob['name'], 'data')
		target_path = os.path.join(output_folder, day_name, prob['name'], '1')
		#remkdir(os.path.join(output_folder, day_name, prob['name']))
		remkdir(target_path)
		for i in range(1, prob['test cases'] + 1):
			for suf in ['.in', '.ans']:
				copy(data_path, prob['name'] + str(i) + suf, target_path)
		if not copy(os.path.join(data_path, 'chk'), 'chk.cpp', target_path) and 'uoj_builtin_checker' not in prob:
			prob['uoj_builtin_checker'] = 'wcmp'
		if not copy(os.path.join(data_path, 'val'), 'val.cpp', target_path):
			pass
			#with open(os.path.join(target_path, 'val.cpp'), 'w') as f:
			#	f.write(empty_cpp)
		for name in os.listdir(data_path):
			if os.path.join(data_path, name) not in base.copied_data:
				require_path = os.path.join(target_path, 'require')
				if not os.path.exists(require_path):
					os.makedirs(require_path)
				warning('Unusual file \'%s\' found.' % os.path.join(data_path, name))
				copy(data_path, name, require_path)
		download_path = os.path.join(target_path, 'download')
		remkdir(download_path)
		case_no = (0 if prob['type'] == 'output' else prob['sample count'])
		suffices = ['.in', '.ans']
		data_path = os.path.join(day_name, prob['name'], 'down')
		for i in range(1, case_no + 1):
			for suf in suffices:
				#copy(data_path, prob['name'] + str(i) + suf, download_path)
				copy(data_path, prob['name'] + str(i) + suf, os.path.join(target_path, 'ex_' + prob['name'] + str(i) + suf))
		copy(data_path, 'checker', download_path)
		for name in os.listdir(data_path):
			if os.path.join(data_path, name) not in base.copied_data:
				warning('Unusual file \'%s\' found.' % os.path.join(data_path, name))
				copy(data_path, name, download_path)
		with open(os.path.join(target_path, 'problem.conf'), 'w') as f:
			f.write(prob2uoj_conf(prob))
		#with open(os.path.join(target_path, 'std.cpp'), 'w') as f:
		#	f.write(empty_cpp)
		if os.system('dos2unix %s 2> log' % os.path.join(output_folder, day_name, prob['name'], '1', '*')) != 0:
			warning('dos2unix failed.')
	if os.path.exists('log'):
		os.remove('log')

def copy_files(suffix = ''):
	for day in base.days():
		base.mkdir(os.path.join(output_folder, day.route))
	for prob in base.probs():
		eval(suffix + 'copy_problem_files')(prob)

def zip_tree(z, path):
	for name in os.listdir(path):
		full_path = os.path.join(path, name)
		if os.path.isdir(full_path):
			zip_tree(z, full_path)
		else:
			z.write(full_path)

def make_zip():
	for day_name, day_data in base.probs.items():
		if day_name not in base.day_set:
			continue
		with zipfile.ZipFile(os.path.join(output_folder, day_name + '.zip'), 'w') as z:
			zip_tree(z, os.path.join(output_folder, day_name))

def noi():
	global output_folder
	base.no_compiling = False
	output_folder = 'noi'
	remkdir('noi')
	copy_files()

def test():
	global output_folder
	base.no_compiling = False
	output_folder = 'bin'
	if base.conf.folder != 'problem':
		remkdir('bin')
		for day in base.days():
			remkdir(base.pjoin('bin', day.route))
	copy_files('test_')
	
def pc2():
	global output_folder
	base.no_compiling = True
	remkdir('pc2')
	copy_files('pc2_')

def release():
	global output_folder
	infom('make release files.')
	remkdir('release')
	base.no_compiling = True
	output_folder = 'release'
	copy_files()
	make_zip()

def svn_init():
	for day_name, probs in base.probs.items():
		if day_name not in base.day_set:
			continue
		for prob in probs:
			if base.rjoin(day_name, prob['name']) not in base.prob_set:
				continue
			if not os.path.exists(os.path.join('uoj', day_name, prob['name'], '.svn')):
				os.system(
					'svn checkout --username %s --password %s %s %s' % (
						uoj_config['user'],
						uoj_config['svn pwd'],
						'svn://%s/problem/%d' % (uoj_config['host'], prob['uoj id']),
						os.path.join(os.path.join('uoj', day_name, prob['name']))
					)
				)
			else:
				os.system('svn delete %s' % os.path.join(os.path.join('uoj', day_name, prob['name'], '1')))
	
def svn_upload():
	for day_name, probs in base.probs.items():
		if day_name not in base.day_set:
			continue
		for prob in probs:
			if base.rjoin(day_name, prob['name']) not in base.prob_set:
				continue
			os.system('svn add %s' % os.path.join(os.path.join('uoj', day_name, prob['name'], '1')))
			os.system('svn commit %s -m "%s"' % (
				os.path.join(os.path.join('uoj', day_name, prob['name'], '1')),
				'Auto generated by oi_tools'
			))

def uoj():
	global output_folder, uoj_config
	base.no_compiling = False
	output_folder = 'uoj'
	if not os.path.exists('uoj'):
		os.makedirs('uoj')
	if not os.path.exists('uoj.json'):
		flag = False
		warning('uoj.json not found. Cannot upload to uoj')
	else:
		flag = True
	if flag:
		uoj_config = json.loads(open('uoj.json').read())
		svn_init()
	copy_files('uoj_')
	if flag:
		svn_upload()

work_list = {
	'noi' : noi,
	'test' : test,
	'uoj' : uoj,
	'pc2' : pc2,
	'release' : release
}

if __name__ == '__main__':
	if base.init():
		infom('Packing starts at %s.\n' % str(datetime.datetime.now()))
		for base.work in base.works:
			work_list[base.work]()
	else:
		print('Use arguments other than options to run what to output.')
		print('Enabled output types:')
		print('\ttest: Generate output for test. Must run this output before tester.py.')
		print('\trelease: Generate what is suitable to publish.')
		print('\tnoi: Generate output in noi style.')
		print('\tuoj: Generate output in uoj style.')



# -*- coding: utf-8 -*-

from __future__ import print_function
import os, re, json, shutil, time
from . import base
from .base import log, pjoin

def tsinsen_oj():
	new_tc = []
	
	class Base(object):
		def __init__(self):
			pass
		def write(self, text):
			pass
		def close(self):
			pass
	class Conf(Base):
		def __init__(self, *names):
			self.names = names
			self.buff = b''
		def write(self, text):
			self.buff += text
		def close(self):
			if type(self.buff) == bytes:
				cur = self.buff.decode('utf-8').strip()
			else:
				cur = self.buff
			for name in reversed(self.names[1:]):
				if not name:
					cur = [cur]
				else:
					cur = {name : cur}
			base.conf[self.names[0]] = cur
	class File(Base):
		def __init__(self, *names):
			path = ''
			for name in names[:-1]:
				path = base.pjoin(path, name)
				if not os.path.exists(path):
					os.makedirs(path)
			self.f = open(base.pjoin(*names), 'wb')
		def write(self, text):
			self.f.write(text)
		def close(self):
			self.f.close()
	class HtmlFile(Base):
		def __init__(self, *names):
			path = ''
			for name in names[:-1]:
				path = base.pjoin(path, name)
				if not os.path.exists(path):
					os.makedirs(path)
			self.fname = base.pjoin(*names)
			self.buff = b''
		def write(self, text):
			self.buff += text.strip() + b'\n'
		def close(self):
			ret = json.dumps(json.dumps(self.buff.decode('utf-8'))).encode('utf-8')
			f = open(self.fname, 'wb')
			f.write(b'{{ self.title() }}\n')
			f.write(b'{{ render(%s, \'html\') }}\n' % ret)
			f.close()
	class FileName(Base):
		last_name = None
		def write(self, text):
			FileName.last_name = text.decode('utf-8').rstrip('\n\r')
	class Base64File(Base):
		def __init__(self):
			self.buff = b''
		def write(self, text):
			self.buff += text
		def close(self):
			import base64
			if not os.path.exists('resources'):
				os.makedirs('resources')
			open(base.pjoin('resources', FileName.last_name), 'wb').write(base64.b64decode(self.buff))
	Title = lambda : Conf('title', 'zh-cn')
	CheckPoint = lambda : Conf('key words')
	TestMethod = Base
	InputFileName = Base
	OutputFileName = Base
	Description = lambda : HtmlFile('statement', 'zh-cn.md')
	Judger = lambda : File('data', 'chk', 'chk.cpp')
	
	class TimeLimit(Conf):
		def __init__(self):
			super(TimeLimit, self).__init__('time limit')
		def close(self):
			self.buff = float(self.buff.strip()[:-1])
			super(TimeLimit, self).close()
	
	class MemoryLimit(Conf):
		def __init__(self):
			super(MemoryLimit, self).__init__('memory limit')
		def close(self):
			if len(self.buff) >= 2:
				s = self.buff.strip().decode('utf-8')
				sp = 1 if '0' <= s[-2] <= '9' else 2
				self.buff = (s[:-sp] + ' ' + s[-sp:]).encode('utf-8')
			super(MemoryLimit, self).close()
	
	class Solution(File):
		def __init__(self):
			super(Solution, self).__init__('tsinsen-oj', 'std', 'std.cpp')
		def close(self):
			super(Solution, self).close()
			if 'users' not in base.conf:
				base.conf['users'] = {}
			if 'tsinsen-oj' not in base.conf['users']:
				base.conf['users']['tsinsen-oj'] = {}
			base.conf['users']['tsinsen-oj']['std'] = 'tsinsen-oj/std/std.cpp'
			
	class InData(File):
		last_in = None
		def __init__(self):
			tc = set(base.conf.test_cases)
			for i in range(1, len(tc) + 2):
				if str(i) not in tc:
					name = i
			new_tc.append(name)
			base.conf.test_cases.append(name)
			InData.last_in = name
			super(InData, self).__init__('data', '%d.in' % name)
			
	class OutData(File):
		def __init__(self):
			super(OutData, self).__init__('data', '%d.ans' % InData.last_in)
			
	class Checkers(File):
		fname = ('data', 'chk', 'chk.cpp')
		def __init__(self):
			super(Checkers, self).__init__(*Checkers.fname)
		def close(self):
			super(Checkers, self).close()
			if os.path.getsize(base.pjoin(*Checkers.fname)) < 10:
				base.rmtree(base.pjoin(*Checkers.fname[:2]), ignore_errors = True)
	import re
	key_re = re.compile('(\w*)\((\w*)\)')

	if base.conf.folder != 'problem':
		log.error(u'只能导入到一个problem的工程中。')
		return
	status = None
	for line in open(base.args[0], 'rb'):
		if not status:			# 啥都没有
			try:
				key = line.decode('utf-8').strip()[:-1]
				m = key_re.match(key)
				if m:
					if m.group(1) == 'File':
						log.warning(u'题面含有文件，请在题面中手工查找字符串`%s`并进行相应修改，该文件保存在`resources/%s`。' % (
							m.group(2), FileName.last_name
						))
						buffer = Base64File()
					else:
						log.warning(u'无法解析关键词`%s`。' % key)
				else:
					buffer = eval(key)()
			except Exception as e:
				log.warning(u'无法解析关键词`%s`。' % key)
				log.debug(e)
				buffer = Base()
			status = True
		elif status == True:	# 读到了等号的行，还没读到标记开始的行
			status = line
		elif status == line:	# 读到了标记结束的行
			buffer.close()
			status = None
		else:					# 读到了标记开始的行，还没读到标记结束的行
			buffer.write(line)
	if len(new_tc) > 0:
		if 'data' not in base.conf:
			base.conf['data'] = []
		base.conf['data'].append({'cases' : new_tc})
	base.save_json(base.conf)

def syzoj(style = 'loj', cookies = None):
	if base.conf.folder != 'problem':
		log.error(u'只能导入到一个problem的工程中。')
		return
	import requests
	try:
		headers = {'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.102 Safari/537.36'}
		r = requests.get(base.args[0] + '/export', headers = headers, cookies = cookies)
		data = json.loads(r.content)
		assert(data['success'])
		data = data['obj']
	except Exception as e:
		log.error(u'下载或解析`' + base.args[0] + '/export`失败。')
		log.info(e)
	conf = base.conf
	conf['time limit'] = float(data.get('time_limit', 1000)) / 1000.
	conf['memory limit'] = str(data.get('memory_limit', 512)) + ' MiB'
	conf['type'] = {'traditional' : 'program', 'submit-answer' : 'output'}.get(data.get('type'), 'interactive')
	conf['key words'] = data.get('tags', [])
	conf.setdefault('title', {})
	conf['title']['zh-cn'] = data.get('title', u'标题读取失败')
	with open('statement/zh-cn.md', 'wb') as f:
		f.write(b'{{ self.title() }}\n\n')
		if style != 'ipuoj':
			for key, name in [
				('description', u'题目描述'),
				('input_format', u'输入格式'),
				('output_format', u'输出格式'),
				('example', u'样例'),
				('limit_and_hint', u'子任务')
			]:
				if key in data:
					f.write(('## %s\n\n' % name).encode('utf-8'))
					f.write(data[key].encode('utf-8'))
					f.write(b'\n\n')
		else:
			f.write(data['description'].encode('utf-8'))
			f.write(b'\n\n')
	def download_file(url, local_filename):
		with requests.get(url, headers = headers, stream = True, cookies = cookies) as r:
			r.raise_for_status()
			with open(local_filename, 'wb') as f:
				for chunk in r.iter_content(chunk_size = 8192): 
					if chunk:
						f.write(chunk)
						f.flush()
	import zipfile
	download_file(base.args[0] + '/testdata/download', 'data.zip')
	with zipfile.ZipFile('data.zip', 'r') as z:
		z.extractall('data')
		for name in os.listdir('data'):
			if name.endswith('.out'):
				if os.path.exists(pjoin('data', name[:-4] + '.ans')):
					os.remove(pjoin('data', name[:-4] + '.ans'))
					time.sleep(.1)
				os.rename(pjoin('data', name), pjoin('data', name[:-4] + '.ans'))
	if data.get('have_additional_file'):
		download_file(base.args[0] + '/download/additional_file', 'down.zip')
		with zipfile.ZipFile('down.zip', 'r') as z:
			z.extractall('down')
	else:
		base.rmtree('down')
		time.sleep(.1)
		os.makedirs('down')
	base.rmtree('pre')
	time.sleep(.1)
	os.makedirs('pre')
	base.save_json(base.conf)
	from . import gen
	base.run_exc(gen.work_list['auto'])

def ipuoj():
	tool_conf = base.tool_conf['ipuoj']['default']
	cookies = tool_conf['cookies']
	syzoj('ipuoj', cookies)

def loj():
	syzoj('loj')

def data_folder(bpath = None, data_type = 'data'):
	fd = {
		'data' : 'data',
		'samples' : 'down'
	}.get(data_type)
	if base.conf.folder != 'problem':
		log.error(u'只能导入到一个problem的工程中。')
		return
	if bpath is None:
		bpath = base.args[0]
	cases = set(base.conf.test_cases if data_type == 'data' else base.conf.sample_cases)
	idx = [1]	# python2居然不支持nonlocal，只能用这种鬼畜方法实现闭包了
	while idx[0] in cases or str(idx[0]) in cases:
		idx[0] += 1
	def save(path, iname, oname):
		log.info(u'找到数据`%s/{%s,%s}`' % (path, iname, oname))
		shutil.copyfile(pjoin(path, iname), pjoin(base.conf.path, fd, str(idx[0]) + '.in'))
		shutil.copyfile(pjoin(path, oname), pjoin(base.conf.path, fd, str(idx[0]) + '.ans'))
		idx[0] += 1
	re_inp = re.compile('(.*)input(.*)')
	re_in = re.compile('(.*)in(.*)')
	re_e = re.compile('(.*)()')
	def scan(path):
		for fname in base.sorter()(list(os.listdir(path))):
			flag = False
			fpath = pjoin(path, fname)
			if os.path.isfile(fpath):
				for re_i, ostrs in ((re_inp, ('answer', 'output')), (re_in, ('ans', 'out')), (re_e, ('.a', ))):
					ma = re_i.match(fname)
					if ma:
						for ostr in ostrs:
							oname = ma.group(1) + ostr + ma.group(2)
							if os.path.exists(pjoin(path, oname)):
								save(path, fname, oname)
								flag = True
								break
						if flag:
							break
			else:
				scan(fpath)
	scan(bpath)
	from . import gen
	base.run_exc(gen.work_list[data_type])

def data_pack(suf, bpath = None, data_type = 'data'):
	import zipfile, tarfile, rarfile
	packer = {
		'zip' : zipfile.ZipFile,
		'rar' : rarfile.RarFile,
		'tar' : tarfile.TarFile
	}.get(suf, None)
	if suf == 'rar':
		base.check_install('unrar')
	if not packer:
		log.error(u'暂时不支持的压缩包类型。')
		return
	if base.conf.folder != 'problem':
		log.error(u'只能导入到一个problem的工程中。')
		return
	if bpath is None:
		bpath = base.args[0]
	with packer(bpath) as z:
		z.extractall('tmp')
	data_folder('tmp', data_type)
	time.sleep(.1)
	base.rmtree('tmp')

def data(bpath = None, data_type = 'data'):
	if base.conf.folder != 'problem':
		log.error(u'只能导入到一个problem的工程中。')
		return
	if bpath is None:
		bpath = base.args[0]
	suffix = [
		('.rar', 'rar'),
		('.zip', 'zip'),
		('.tar.gz', 'tar'),
		('.tar', 'tar')
	]
	if os.path.isfile(bpath):
		for suf, tp in suffix:
			if bpath.endswith(suf):
				data_pack(tp, bpath, data_type)
				break
		else:
			log.error(u'暂时不支持的压缩包类型。')
	else:
		data_folder(bpath, data_type)

work_list = {
	'tsinsen-oj' : tsinsen_oj,
	'loj' : loj,
	'ipuoj' : ipuoj
}

work_list['data'] = lambda bp = None: data(bp, 'data')
work_list['data-folder'] = lambda bp = None: data_folder(bp, 'data')
for z in ['zip', 'tar', 'rar']:
	work_list['data-' + z] = lambda bp = None: data_pack(z, bp, 'data')
work_list['samples'] = lambda bp = None: data(bp, 'samples')
work_list['samples-folder'] = lambda bp = None: data_folder(bp, 'samples')
for z in ['zip', 'tar', 'rar']:
	work_list['samples-' + z] = lambda bp = None: data_pack(z, bp, 'samples')

if __name__ == '__main__':
	try:
		if base.init() and len(base.args) != 0:
			for base.work in base.works:
				base.run_exc(work_list[base.work])
		else:
			log.info(u'这个工具用于导入其他类型的工程，参数1必须是来源路径。')
			#log.info(u'支持的工作：%s' % ','.join(sorted(work_list.keys())))
			log.info(u'支持的工作：')
			log.info(u'  tsinsen-oj  参数1是符合清澄OJ“我来出题”的上传下载格式的本地文件。')
			log.info(u'  loj         参数1是SYZ系OJ的题目网站如“https://loj.ac/problem/1”。')
			log.info(u'  data        参数1是一个装数据的文件夹或压缩包，只导入数据。')
			log.info(u'  data-folder 在上述命令不好使的时候，强制指定是文件夹。')
			log.info(u'  data-zip    在上述命令不好使的时候，强制指定是zip文件。')
			log.info(u'  data-tar    在上述命令不好使的时候，强制指定是tar文件。')
			log.info(u'  data-rar    在上述命令不好使的时候，强制指定是rar文件。')
			log.info(u'  samples*    把上述命令的data换成samples用于处理样例。')
	except base.NoFileException as e:
		log.error(e)
		log.info(u'尝试使用`python -m tuack.gen -h`获取如何生成一个工程。')



help = u'''\
用于造数据，把gen.py放在problem目录下，然后运行：
python -m tuack.dmk [开关]
具体写法在gen.py有相关说明。
-s i,compile,g,run  跳过若干个步骤不执行。
                    i,init：初始化函数
                    c,compile：编译标程
                    g,gen：生成数据
                    r,run：运行标程
'''

from .base import log
from . import base
import yaml, os, time, sys, random

folders = ['down', 'data']
fmap = {'down' : 'samples', 'data' : 'data'}

class Getter:
	def __init__(self, arg1, arg2):
		self.arg1 = arg1
		self.arg2 = arg2
	def __getitem__(self, key):
		return self.arg1.get(key, self.arg2.get(key, None))

def main(argv):
	base.init()
	skip = set()
	flag = None
	for arg in argv:
		if arg == '-h':
			log.info(help)
			return
		elif flag == '-s':
			for key in arg.split(','):
				skip.add({
					'i' : 'init',
					'c' : 'compile',
					'g' : 'gen',
					'r' : 'run'
				}.get(key, key))
			flag = None
		elif flag:
			log.error(f"错误的参数{flag}。")
			return
		elif arg.startswith('-'):
			flag = arg
		else:
			log.info(f"未知的传入信息{arg}。")
	import gen, platform
	if 'init' not in skip:
		if hasattr(gen, 'init'):
			gen.init()
		else:
			log.info('没有init函数，跳过执行。')
	conf = base.conf
	if 'compile' not in skip:
		if not hasattr(gen, 'std'):
			log.info('没有std变量，将查找第一个expect能取到100的标程。')
			gen_std = None
			for user, codes in conf.get('users').items():
				for code, info in codes.items():
					expc = info.get('expected', '')
					if type(expc) == str:
						expc = [expc]
					for e in expc:
						if eval('100 ' + e) == True:
							gen_std = (user, code)
							break
					else:
						continue
					break
				if gen_std: break
			if not gen_std: log.warning('没有找到任何可能像标程的代码，请配置conf.yaml的users字段。')
		else:
			gen_std = gen.std
		if gen_std:
			os.system('g++ %s -o std.exe -std=c++17 -O2 -Wall -Wextra %s' % (
				conf.get('users')[gen_std[0]][gen_std[1]].get('path'),
				f"-Wl,--stack={int(conf.ml().B)}" if platform.system() == 'Windows' else ''
			))
	std_exe = False
	if 'run' not in skip:
		if os.path.isfile('std.exe'):
			std_exe = True
		else:
			log.warning('没有std.exe，无法生成输出文件。')
	if os.path.isfile('random_state'):
		random.setstate(eval(open('random_state').read()))
	else:
		open('random_state', 'w').write(str(random.getstate()))
	warnend_gen_gen = False
	for folder in folders:
		os.makedirs(folder, exist_ok=True)
		for item in conf[fmap[folder]]:
			for case in item['cases']:
				log.info(f'数据类型{folder}，测试点{case}。')
				if hasattr(gen, 'gen'):
					if 'gen' not in skip and not item['args'].get('manual'):
						with open(f'{folder}/{case}.in', 'w') as f:
							gen.gen(
								f,
								Getter(item.get('args', {}), conf.get('args', {})),
								{
									'case' : case,
									'folder' : folder
								}
							)
				elif not warnend_gen_gen:
					log.warning('没有数据生成函数gen，跳过数据生成。')
					warnend_gen_gen = True
				if std_exe:
					t = time.perf_counter()

					if base.system == 'Windows':
						os.system(f'std.exe < {folder}/{case}.in > {folder}/{case}.ans')
					else:
						os.system(f'./std.exe < {folder}/{case}.in > {folder}/{case}.ans')

					log.info(f"标程用时：{time.perf_counter() - t}")
					ans = ' '.join([line.strip() for line in open(f'{folder}/{case}.ans')])
					if len(ans) > 100: 
						ans = ans[:100] + f"（省略{len(ans) - 100}个字符。）"
					log.info('答案：' + ans)

if __name__ == '__main__':
	main(sys.argv[1:])



# -*- coding: utf-8 -*-

from __future__ import print_function
import os
import re
import sys
import json
import datetime
import shutil
import subprocess
import time
import signal
import zipfile
from multiprocessing import Process, Queue
from functools import wraps
from threading import Timer
import platform
import gettext
import yaml
from . import base
from .base import log, pjoin, tmp_path
try:
	import jinja2
except:
	pass
from . import tools
import uuid
import traceback

work_class = {
	'noi' : {'noi'},
	'ccpc' : {'ccpc'},
	'uoj' : {'uoj'},
	'loj' : {'loj'},
	'ipuoj' : {'ipuoj'},
	'tupc' : {'tupc'},
	'tuoj-pc' : {'tupc', 'tuoj'},
	'tuoi' : {'tuoi'},
	'tuoj-oi' : {'tuoi', 'tuoj'},
	'tuoj' : {'tuoj'},
	'thuoj' : {'thuoj'},
	'hand' : {'hand'},
	'ccc' : {'ccc-tex', 'ccc-md'},
	'ccc-tex' : {'ccc-tex'},
	'ccc-md' : {'ccc-md'},
	'tsinsen-oj' : {'tsinsen-oj'},
	'tex' : {'noi', 'ccpc', 'tupc', 'tuoi', 'ccc-tex', 'hand'},
	'md' : {'uoj', 'tuoj', 'ccc-md', 'loj', 'ipuoj'},
	'html' : {'tsinsen-oj'},
	'doku' : {'thuoj'},
	'syzoj_like' : {'loj', 'ipuoj'}
}
io_styles = {
	'noi' : 'fio',
	'ccpc' : 'stdio',
	'uoj' : 'stdio',
	'tuoj' : 'stdio',
	'ccc' : 'stdio',
	'tuoi' : 'stdio',
	'tupc' : 'stdio',
	'tsinsen-oj' : 'stdio',
	'loj' : 'stdio',
	'ipuoj' : 'stdio',
	'thuoj' : 'stdio',
	'hand' : 'stdio'
}
base_templates = {
	'noi' : 'tuoi',
	'tuoi' : 'tuoi',
	'tupc' : 'tupc',
	'ccpc' : 'ccpc',
	'ccc' : 'ccc',
	'hand' : 'hand'
}

secondary_dict = {}
env = None
pandoc_version = None

def detect_pandoc_version():
	global pandoc_version
	if pandoc_version is not None:
		return
	cp = subprocess.Popen(['pandoc', '-v'], stdout = subprocess.PIPE)
	line = cp.stdout.read().decode("utf-8", "ignore").splitlines()[0]
	pandoc_version = tuple(map(int, line.split()[-1].split('.')))

def get_pandoc_option():
	detect_pandoc_version()
	if pandoc_version >= (2, ):
		option = '--listings -f markdown-smart -t latex-smart'
	else:
		option = '--no-tex-ligatures --listings -t latex'
	return option

def get_template(fname, lang = None):
	global env
	if base.lang:
		lang = base.lang
	if not lang:
		lang = 'zh-cn'
	if env == None or base.system == 'Darwin' or (lang and env['lang'] != lang):
		env = {
			'env' : jinja2.Environment(
				loader = jinja2.FileSystemLoader(tmp_path),
				extensions = ['jinja2.ext.do', 'jinja2.ext.i18n']
			),
			'lang' : lang
		}
		try:
			tra = gettext.translation('lang', tmp_path, [env['lang']])
			env['env'].install_gettext_translations(tra)
		except FileNotFoundError as e:
			log.warning(u'没有%s这种语言的翻译表可用。' % lang)
	return env['env'].get_template(fname)

def table(path, name, temp, context, options):
	if options == None:
		options = {}
	for suf in ['.py', '.pyinc', '.json', '.yaml', '.yml']:
		try:
			base.copy(path, name + suf, pjoin(tmp_path, 'table' + suf))
			log.info(u'渲染表格`%s`，参数%s' % (base.pjoin(path, name + suf), str(options)))
			break
		except base.NoFileException as e:
			pass
	else:
		raise base.NoFileException(u'找不到表格`%s.*`' % base.pjoin(path, name))
	def merge_ver(table, titled = None):
		if titled is not None:
			options['titled'] = titled
		return table
	def merge(table):
		titled = options.get('titled', True)
		ret = [row for row in table]
		last = ret[-1]
		for i in range(len(table) - 2, -1, -1):
			for j in range(len(last)):
				cur = ret[i][j]
				if type(cur) == tuple:
					cur = cur[0]
				las = last[j]
				if type(las) == tuple:
					las, opt = las
				else:
					opt = {}
				last[j] = las
				if (not titled or i != 0) and las == cur and not opt.get('no_merge') and options.get('merge') != False:
					last[j] = None
				if (opt.get('no_merge') or options.get('merge') == False) and context.get('comp') in work_class['syzoj_like']:
					last[j] = str(las) + f"<!--{ uuid.uuid4() }-->"
			last = ret[i]
		if len(table) > 0:
			for j in range(len(table[0])):
				if type(table[0][j]) == tuple:
					table[0][j] = table[0][j][0]
		if titled and len(table) > 1:
			for j in range(len(table[1])):
				if type(table[1][j]) == tuple:
					table[1][j] = table[1][j][0]
		return ret
	if suf == '.json':
		res = get_template('table.json', context['prob'].lang()).render(context, options = options)
		try:
			table = json.loads(res)
		except Exception as e:
			open(pjoin(tmp_path, 'table.tmp.json'), 'w').write(res)
			log.error(u'json文件错误`tmp/table.tmp.json`')
			raise e
		os.remove(pjoin(tmp_path, 'table.json'))
	elif suf == '.yaml' or suf == '.yml':
		try:
			table = yaml.full_load(open(base.pjoin(tmp_path, 'table' + suf), 'rb').read().decode('utf-8'))
		except Exception as e:
			open(pjoin(tmp_path, 'table.tmp.json'), 'w').write(res)
			log.error(u'json文件错误`tmp/table.tmp.json`')
			raise e
	elif suf == '.py' or suf == '.pyinc':
		no_merge = lambda item : (item, {'no_merge' : True})
		code = 'def render(context, options, merge_ver, no_merge):\n'
		code += '\tglobal val\n'
		code += '\tfor key, val in context.items():\n'
		code += '\t\texec(\'%s = val\' % key, globals())\n'
		for line in open(base.pjoin(tmp_path, 'table' + suf), 'rb'):
			code += '\t' + line.decode('utf-8').rstrip() + '\n'
		namespace = {}
		exec(code, namespace)
		try:
			table = namespace['render'](context, options, merge_ver, no_merge)
		except Exception as e:
			log.error(e)
			lines = traceback.format_exc().splitlines()
			for idx, line in enumerate(lines):
				if '<string>' in line:
					m = re.match(r'^.*line (\d*),.*$', line)
					if m:
						lineno = int(m.group(1)) - 4
						log.info('  File "%s", line %d' % (base.pjoin(path, name + suf), lineno))
						log.info(code.splitlines()[lineno + 3][1:])
						for i in range(idx + 1, len(lines)):
							log.info(lines[i])
						break
			else:
				raise e
			raise e
	table = merge(table)
	table = json.loads(json.dumps(table))
	if context.get('comp') in work_class['syzoj_like'] and len(table) > 0:
		last_line = [None] * len(table[0])
		for i in range(len(table)):
			for j in range(len(table[i])):
				if table[i][j] == None:
					table[i][j] = last_line[j]
				else:
					last_line[j] = table[i][j]
	cnt = [[1] * len(row) for row in table]
	max_len = len(table[-1])
	for i in range(len(table) - 2, -1, -1):
		max_len = max(max_len, len(table[i]))
		for j in range(min(len(table[i]), len(table[i + 1]))):
			if table[i + 1][j] == None:
				cnt[i][j] += cnt[i + 1][j]
	ret = get_template(temp, context['prob'].lang()).render(context, table = table, cnt = cnt, width = max_len, options = options)
	return ret

def to_arg(dic):
	return ','.join(['%s = %s' % (key, val if type(val) != str else json.dumps(val)) for key, val in dic.items()])

def uoj_title(text):	## 很显然，这个是在瞎整，不过UOJ为什么从二级标题开始用？修改标题字体大小应该改CSS而不是题目嘛
	in_quote = 0
	result = []
	for line in text.split(b'\n'):
		if in_quote == 0 and line.startswith(b'#'):
			result.append(b'#' + line)
		else:
			result.append(line)
		in_quote ^= line.count(b'```') & 1
	return b'\n'.join(result)
	
def loj_bug(text):		## 学弟也学着我瞎整，伤心QAQ
	# TODO: Avoid duplication
	in_quote = 0
	result = []
	for line in text.split(b'\n'):
		if in_quote == 0 and line.startswith(b'<table'):
			result.append(line
				.replace(b'$<', b'$ <')   # LOJ Markdown 渲染的已知 bug
			)
		else:
			result.append(line)
		in_quote ^= line.count(b'```') & 1
	return b'\n'.join(result)

class Base(object):
	def __init__(self, comp, conf = None):
		self.conf = (base.conf if conf == None else conf)
		self.comp = comp
		self.day = None
		self.contest = None
		self.prob = None
		self.path = None
		self.prec = None
		self.secondary_dict = {}
		self.io_style = io_styles[comp]

	def file_name(self, name):
		if self.comp == 'noi':
			return self.prob['name'] + '/' + name
		elif self.comp == 'uoj':
			return name
		else:
			return name

	def secondary(self, s, sp = None):
		if sp != None:
			if type(sp) == str:
				sp = [sp]
			in_set = False
			for i in sp:
				if i in work_class and base.work in work_class[i]:
					in_set = True
					break
			if not in_set:
				return ''
		if self.work == 'tex' or self.work == 'html' or self.work == 'doku':
			id = str(uuid.uuid4())
			self.secondary_dict[id] = s
			return id
		else:
			return ' {{ ' + s + ' }} '

	def resource(self, name):
		if self.comp in work_class['syzoj_like'] and self.prob.get('pid', {}).get(self.comp + '-default'):
			return '/../problem/show_image/%d/' % self.prob.get('pid', {}).get(self.comp + '-default') + name
		return self.prob['name'] + '/' + name
		
	def down_file(self, name):
		ret = ''
		fname = pjoin(self.prob.path, 'down', name)
		length_warned = False
		for idx, line in enumerate(open(fname, 'rb')):
			if len(line) > 60 and not length_warned:
				log.warning(u'文件`%s`的第%d行太长，建议只提供下发而不渲染到题面。' % (fname, idx + 1))
				length_warned = True
			ret += line.decode('utf-8')
		return ret

	def hide_file(self, name):
		ret = ''
		fname = pjoin(self.prob.path, 'hide', name)
		length_warned = False
		for idx, line in enumerate(open(fname, 'rb')):
			if len(line) > 60 and not length_warned:
				log.warning(u'文件`%s`的第%d行太长，建议只提供下发而不渲染到题面。' % (fname, idx + 1))
				length_warned = True
			ret += line.decode('utf-8')
		if self.comp not in work_class['tex']:
			if ret < 20:
				return u'隐藏内容'
			c = len(ret) // 20
			ret =  ret[:c] + u'……隐藏内容……' + ret[-c:]
		return ret
		
	def titlize(self, title, args, lang = None):
		if base.lang:
			lang = base.lang
		if not lang:
			lang = 'zh-cn'
		if title == 'sample':
			if len(args) > 0:
				self.context['vars']['sample_id'] = args[0]
			if len(args) <= 1:
				return ''
		tra = gettext.translation('lang', tmp_path, [env['lang']])
		ret = tra.gettext(title)
		if len(args) > 0:
			try:
				ret = ret % args
			except Exception as e:
				ret = ret + ''.join(map(str, args))
		elif title == 'sample explanation':
			try:
				ret = ret % (self.context['vars'].get('sample_id', ''), )
			except Exception as e:
				pass
		else:
			try:
				ret = ret % ('',)
			except Exception as e:
				pass
		return '## ' + ret

	def ren_prob_md_j(self):
		time.sleep(0.5)
		log.info(u'渲染题目题面 %s %s' % (self.comp, self.prob.route))
		try:
			shutil.copy(self.prob.statement(), pjoin(tmp_path, 'problem.md.jinja'))
			time.sleep(0.1)
		except base.NoFileException as e:
			log.error(u'找不到题面文件，建议使用`python -m tuack.gen problem`生成题目工程。')
			return
		def to_I(x):
			if not issubclass(type(x), dict) and not issubclass(type(x), list): return
			for key in (x if issubclass(type(x), dict) else range(len(x))):
				if type(x[key]) == int:
					x[key] = tools.Int(x[key])
				elif type(x[key]) == float:
					x[key] = tools.Float(x[key])
				else:
					to_I(x[key])
		to_I(self.prob)
		self.context = {
			'prob' : self.prob,
			'args' : self.prob.get('args'),
			'data' : self.prob.get('data'),
			'samples' : self.prob.get('samples'),
			'pre' : self.prob.get('pre'),
			'dargs' : tools.a(self.prob.get('data')),
			'pargs' : tools.a(self.prob.get('pre')),
			'sargs' : tools.a(self.prob.get('samples')),
			'aargs' : tools.a(self.prob, self.prob.get('data'), self.prob.get('pre'), self.prob.get('samples'), [self.prob]),
			'day' : self.day,
			'contest' : self.contest,
			'io_style' : self.io_style,
			'comp' : self.comp,
			'tools' : tools,
			'tl' : tools,
			'base' : base,
			'common' : base,
			'file_name' : self.file_name,
			'down_file' : self.down_file,
			'resource' : self.resource,
			'render' : self.secondary,
			'precautions' : self.prec,
			'json' : json,
			'vars' : {},
			's' : lambda title, *args : self.titlize(title, args, self.prob.lang()),
			'input_file_name': None,
			'output_file_name': None,
			'work_class' : work_class,
			'noi_pas_c': base.noi_pas_c
		}
		if self.prob.get('file io'):
			self.context['input_file_name'] = self.prob.get('input_table', {}).get('zh-cn', self.prob['name'] + '.in')
			self.context['output_file_name'] = self.prob.get('output_table', {}).get('zh-cn', self.prob['name'] + '.out')
		self.context['img'] = lambda src, env = None, **argw : self.context['render'](
			"template('image', resource = resource(%s), %s)" % (json.dumps(src), to_arg(argw)),
			env
		)
		self.context['font'] = lambda txt, env = None, **argw : self.context['render'](
			"template('font', txt = %s, %s)" % (json.dumps(txt), to_arg(argw)),
			env
		)
		self.context['tbl'] = lambda src, env = None, **argw : self.context['render'](
			"table(%s, %s)" % (json.dumps(src), argw),
			env
		)
		if self.comp == 'tsinsen-oj' and 'tsinsen_files' in dir(self.prob):
			self.context['resource'] = lambda name : '/RequireFile.do?fid=%s' % self.prob.tsinsen_files[name]
		open(pjoin(tmp_path, 'problem.md'), 'wb') \
			.write(get_template('problem_base.md.jinja', self.prob.lang())
				.render(self.context)
				.encode('utf-8')
			)
		time.sleep(0.1)

	def check_install(self):
		pass

	def before(self):
		base.mkdir('statements')
		shutil.copytree(pjoin(base.path, 'templates'), tmp_path, dirs_exist_ok = True)
		base.mkdir(pjoin('statements', self.comp))
		if self.conf.folder == 'contest':
			self.contest = self.conf
		elif self.conf.folder == 'day':
			self.day = self.conf
		elif self.conf.folder == 'problem':
			self.prob = self.conf

	def move_resource(self):
		if os.path.exists(pjoin(self.prob.path, 'resources')):
			base.rmtree(self.path, ignore_errors = True)
			time.sleep(0.1)
			shutil.copytree(pjoin(self.prob.path, 'resources'), (self.path if self.prob.route != '' else pjoin(self.path, self.prob['name'])))

	def gen_paths(self):
		self.path = pjoin('statements', self.comp)
		if self.prob.route != '':
			self.path = pjoin(self.path, self.prob.route)
		self.result_path = self.path + ('-' + base.lang if base.lang else '')
		if self.prob.route == '':
			self.result_path = pjoin(self.result_path, self.prob['name'])
		self.result_path += '.' + self.work

	def main(self):
		if self.conf.folder == 'contest':
			for self.day in self.conf.days():
				if not os.path.exists(pjoin('statements', self.comp, self.day.route)):
					os.makedirs(pjoin('statements', self.comp, self.day.route))
		if self.comp == 'tsinsen-oj':
			log.warning(u'如果你使用了resource，你可能需要使用dumper才能获得正确上传清橙的文件。')
		for self.prob in self.conf.probs():
			self.gen_paths()
			self.move_resource()
			self.ren_prob_md_j()
			self.ren_prob_rest()
			self.start_file()

	def final(self):
		base.rmtree(tmp_path, ignore_errors = True)

	def run(self):
		self.check_install()
		self.before()
		self.main()
		self.final()

	def start_file(self):
		if base.start_file:
			base.xopen_file(self.result_path)

class Latex(Base):
	work = 'tex'
	day_templates = {
		'noi' : 'tuoi',
		'tuoi' : 'tuoi',
		'tupc' : 'tupc',
		'ccpc' : 'ccpc',
		'ccc' : 'ccc',
		'hand' : 'hand'
	}

	def resource(self, name):
		return pjoin(self.prob.full_path, 'resources', name).replace('\\', '/')

	def check_install(self):
		base.check_install('pandoc')
		base.check_install('xelatex')
		base.check_install('gettext')

	def ren_prec(self):
		if not os.path.exists(base.pjoin(self.conf.path, 'precautions', 'zh-cn.md')):
			return
		context = {
			'io_style' : self.io_style,
			'comp' : self.comp,
			'tools' : tools,
			'tl' : tools,
			'base' : base,
			'json' : json
		}
		base.copy(base.pjoin(self.conf.path, 'precautions'), 'zh-cn.md', tmp_path)
		open(base.pjoin(tmp_path, 'precautions.md'), 'wb') \
			.write(get_template('zh-cn.md').render(context, conf = self.conf).encode('utf-8'))
		os.system('pandoc %s %s -o %s' % (
			get_pandoc_option(),
			pjoin(tmp_path, 'precautions.md'),
			pjoin(tmp_path, 'precautions.tex')
		))
		self.prec = open(pjoin(tmp_path, 'precautions.tex'), 'rb').read().decode('utf-8')

	def ren_day_prec(self):
		if not os.path.exists(base.pjoin(self.day.path, 'precautions', 'zh-cn.md')):
			return
		context = {
			'io_style' : self.io_style,
			'comp' : self.comp,
			'tools' : tools,
			'tl' : tools,
			'base' : base,
			'json' : json
		}
		base.copy(base.pjoin(self.day.path, 'precautions'), 'zh-cn.md', tmp_path)
		open(base.pjoin(tmp_path, 'precautions.md'), 'wb') \
			.write(get_template('zh-cn.md').render(context, conf = self.day).encode('utf-8'))
		os.system('pandoc %s %s -o %s' % (
			get_pandoc_option(),
			pjoin(tmp_path, 'precautions.md'),
			pjoin(tmp_path, 'precautions.tex')
		))
		self.prec = open(pjoin(tmp_path, 'precautions.tex'), 'rb').read().decode('utf-8')

	def ren_prob_tex(self):
		os.system('pandoc %s %s -o %s' % (
			get_pandoc_option(),
			pjoin(tmp_path, 'problem.md'),
			pjoin(tmp_path, 'problem.tex')
		))
		tex = open(pjoin(tmp_path, 'problem.tex'), 'rb').read().decode('utf-8')
		for key, val in self.secondary_dict.items():	## 如果未来发生了性能问题，这里可以先写到一个文件里然后再解析
			open(pjoin(tmp_path, key + '.tex.jinja'), 'wb').write(
				('{{ ' + val + ' }}').encode('utf-8')
			)
			ret = get_template(key + '.tex.jinja', self.prob.lang()).render(
				self.context,
				template = lambda temp_name, **context : get_template(temp_name + '.tex.jinja', self.prob.lang()).render(context),
				table = lambda name, options = None : table(pjoin(self.prob.path, 'tables'), name, 'table.tex.jinja', self.context, options)
			)
			tex = tex.replace(key, ret)
		self.secondary_dict = {}
		return tex

	def gen_compile(self):
		clangs = {l for prob in self.probs for l in prob.get('compile', {}).keys()}
		ret = {}
		for clang in clangs:
			cur = []
			for prob in self.probs:
				if prob['type'] != 'output':
					if clang in prob['compile']:
						cur.append({
							'option' : prob['compile'][clang] + (' -Wl,--stack=%d' % prob.memory_limit().B if clang in {'cpp', 'c'} and base.out_system == 'Windows' else ''),
							'cnt' : 1,
							'use' : True
						})
					else:
						cur.append({'option' : u'不可用', 'cnt' : 1, 'use' : False})
				else:
					cur.append({'option' : 'N/A', 'cnt' : 1, 'use' : False})
			last = []
			for p in cur:
				if len(last) == 0 or p['option'] != last[-1]['option']:
					last.append(p)
				else:
					last[-1]['cnt'] += 1
			ret[clang] = last
		return ret

	def ren_day_pdf(self):
		log.info(u'渲染比赛日题面 %s %s' % (self.comp, self.day.route if self.day else self.conf.route))
		self.context.pop('prob')
		self.context.pop('file_name')
		self.context.pop('down_file')
		self.context.pop('resource')
		self.context.pop('render')
		self.context['probs'] = self.probs
		self.context['problems'] = self.tex_problems
		self.context['compile'] = self.gen_compile()
		self.context['water_mark'] = base.water_mark
		if base.water_mark:
			shutil.copy(base.water_mark, pjoin(tmp_path, 'water_mark.' + base.water_mark.split('.')[-1]))
		all_problem_statement = get_template('%s.tex.jinja' % base_templates[self.comp]).render(self.context)
		try:
			open(pjoin(tmp_path, 'problems.tex'), 'wb') \
				.write(all_problem_statement.encode('utf-8'))
		except Exception as e:
			log.info(u'渲染出错的文件为`tmp/problems.tmp.tex`')
			open(pjoin(tmp_path, 'problems.tmp.tex'), 'wb') \
				.write(all_problem_statement.encode('utf-8'))
			raise e
		log.info(u'开始使用xelatex渲染题面。')
		back = os.getcwd()
		os.chdir(tmp_path)
		os.system('xelatex -interaction=batchmode problems.tex')
		os.system('xelatex -interaction=batchmode problems.tex')
		if not os.path.exists('problems.pdf'):
			log.warning(u'题面渲染失败，尝试使用关闭静默模式渲染。')
			os.system('xelatex problems.tex')
			os.system('xelatex problems.tex')
		if not os.path.exists('problems.pdf'):
			log.error(u'题面渲染失败，请检查`tmp`目录下各文件。')
		os.chdir(back)
		shutil.copy(pjoin(tmp_path, 'problems.pdf'), self.result_path)

	def ren_day(self):
		if self.day:
			day_name = self.day['name']
			self.probs = list(self.day.probs())
		else:
			day_name = u'测试'
			self.probs = [self.prob]
		if len(self.probs) == 0:
			log.info(u'指定目录`%s`下没有题目可渲染。' % (self.day.path if self.day else self.conf.path))
			return
		self.tex_problems = []
		for self.prob in self.probs:
			self.ren_prob_md_j()
			self.tex_problems.append(self.ren_prob_tex())
		self.ren_day_pdf()
		self.start_file()

	def main(self):
		self.ren_prec()
		self.day_template = self.day_templates[self.comp]
		if self.conf.folder != 'problem':
			for self.day in base.days():
				self.result_path = pjoin('statements', self.comp, self.day.name_lang() + '.pdf')
				self.ren_day_prec()
				self.ren_day()
		else:
			self.result_path = pjoin('statements', self.comp, self.conf.name_lang() + '.pdf')
			self.ren_day()

class Markdown(Base):
	work = 'md'
	def ren_prob_rest(self):
		table_suf = 'html' if self.comp not in work_class['syzoj_like'] else 'md'
		result_md = get_template('problem.md', self.prob.lang()).render(
			self.context,
			template = lambda temp_name, **context : get_template(temp_name + '.html.jinja', self.prob.lang()).render(context),
			table = lambda name, options = None : table(pjoin(self.prob.path, 'tables'), name, 'table.'+ table_suf +'.jinja', self.context, options)
		).encode('utf-8')
		if self.comp == 'uoj':
			result_md = uoj_title(result_md)
		elif self.comp in work_class['syzoj_like']:
			result_md = loj_bug(result_md)
		open(self.result_path, 'wb').write(result_md)

class Html(Base):
	work = 'html'
	def check_install(self):
		base.check_install('pandoc')

	def ren_prob_rest(self):
		if self.comp == 'tsinsen-oj':
			with open(pjoin(tmp_path, 'problem.2.md'), 'wb') as f:
				for line in open(pjoin(tmp_path, 'problem.md'), 'rb'):
					if re.match(b'^##[^#]', line):
						line = (u'## 【%s】\n' % line[2:].strip().decode('utf-8')).encode('utf-8')
					f.write(line)
		else:
			base.copy(tmp_path, 'problem.md', pjoin(tmp_path, 'problem.2.md'))
		txt = open(pjoin(tmp_path, 'problem.2.md'), 'rb').read().decode('utf-8')
		for key, val in self.secondary_dict.items():
			txt = txt.replace(key, '{{ ' + val + ' }}')
		open(pjoin(tmp_path, 'problem.3.md'), 'wb').write(
			txt.encode('utf-8')
		)
		open(base.pjoin(tmp_path, 'problem.4.md'), 'wb') \
			.write(get_template('problem.3.md', self.prob.lang())
				.render(
					self.context,
					template = lambda temp_name, **context : get_template(temp_name + '.html.jinja', self.prob.lang()).render(context),
					table = lambda name, options = None : table(pjoin(self.prob.path, 'tables'), name, 'table.html.jinja', self.context, options)
				).encode('utf-8')
			)
		os.system('pandoc %s -o %s' % (
			pjoin(tmp_path, 'problem.4.md'),
			self.result_path
		))

class DoKuWiki(Base):
	work = 'doku'
	def check_install(self):
		base.check_install('pandoc')

	def ren_prob_rest(self):
		os.system('pandoc %s -o %s -w dokuwiki' % (
			pjoin(tmp_path, 'problem.md'),
			pjoin(tmp_path, 'problem.doku')
		))
		txt = open(pjoin(tmp_path, 'problem.doku'), 'rb').read().decode('utf-8')
		for key, val in self.secondary_dict.items():
			txt = txt.replace(key, '{{ ' + val + ' }}')
		open(pjoin(tmp_path, 'problem.2.doku'), 'wb').write(
			txt.encode('utf-8')
		)
		open(self.result_path, 'wb') \
			.write(get_template('problem.2.doku', self.prob.lang())
				.render(
					self.context,
					template = lambda temp_name, **context : '<HTML>' + get_template(temp_name + '.html.jinja', self.prob.lang()).render(context) + '</HTML>',
					table = lambda name, options = None : '<HTML>' + table(pjoin(self.prob.path, 'tables'), name, 'table.html.jinja', self.context, options).strip() + '</HTML>'
				).encode('utf-8')
			)

class_list = {
	'ccpc' : Latex,
	'uoj' : Markdown,
	'tupc' : Latex,
	'tuoi' : Latex,
	'noi' : Latex,
	'tuoj' : Markdown,
	'ccc-tex' : Latex,
	'ccc-md' : Markdown,
	'tsinsen-oj' : Html,
	'loj' : Markdown,
	'ipuoj' : Markdown,
	'thuoj' : DoKuWiki,
	'hand' : Latex
}

comp_list = {
	'ccpc' : 'ccpc',
	'uoj' : 'uoj',
	'tupc' : 'tupc',
	'tuoi' : 'tuoi',
	'noi' : 'noi',
	'tuoj' : 'tuoj',
	'ccc-tex' : 'ccc',
	'ccc-md' : 'ccc',
	'tsinsen-oj' : 'tsinsen-oj',
	'loj' : 'loj',
	'ipuoj' : 'ipuoj',
	'thuoj' : 'thuoj',
	'hand' : 'hand'
}

if __name__ == '__main__':
	try:
		if base.init():
			base.check_install('jinja2')
			for base.work in base.works:
				for base.lang in base.langs:
					base.run_exc(class_list[base.work](comp_list[base.work]).run)
		else:
			log.info(u'这个工具用于渲染题面。')
			log.info(u'支持的工作：%s' % ','.join(sorted(class_list.keys())))
	except base.NoFileException as e:
		log.error(e)
		log.info(u'尝试使用`python -m tuack.gen -h`获取如何生成一个工程。')
